import argparse
import configparser
import json
import logging
from pathlib import Path
import typing

import boxsdk
from fastapi import Form, HTTPException
import uvicorn
from diycrate import utils

from diycrate.cache_utils import r_c
from diycrate.oauth_utils import setup_oauth, store_tokens_callback
from diycrate.log_utils import setup_logger


setup_logger()

crate_logger = logging.getLogger(__name__)


cloud_provider_name = "Box"

app = utils.FastAPI()


@app.get("/auth_url/")
def auth_url(redirect_url: str):
    """

    :return:
    """
    crate_logger.info("auth url retrieved.")

    app.oauth = setup_oauth(r_c, conf_obj, store_tokens_callback)
    return json.dumps(app.oauth.get_authorization_url(redirect_url=redirect_url))


@app.post("/authenticate")
def authenticate_url(code: typing.Annotated[str, Form()]):
    """

    :return:
    """
    crate_logger.info("authentication flow initiated.")
    app.oauth = setup_oauth(r_c, conf_obj, store_tokens_callback)
    auth_code = code
    return [
        el.decode(encoding="utf-8", errors="strict") if isinstance(el, bytes) else el
        for el in app.oauth.authenticate(auth_code=auth_code)
    ]


@app.post("/new_access")
def new_access(
    access_token: typing.Annotated[typing.Optional[str], Form()],
    refresh_token: typing.Annotated[typing.Optional[str], Form()],
):
    """
    Performs refresh of tokens and returns the result
    :return:
    """
    crate_logger.info("generating an access token.")
    app.oauth = setup_oauth(r_c, conf_obj, store_tokens_callback)
    access_token_to_refresh = access_token
    app.oauth._update_current_tokens(str(access_token_to_refresh), str(refresh_token))
    try:
        refresh_response = app.oauth.refresh(access_token_to_refresh)
        str_response = [
            el.decode(encoding="utf-8", errors="strict")
            if isinstance(el, bytes)
            else el
            for el in refresh_response
        ]
        # we've done the work, so let's wipe the temporary state adjustment clean
        app.oauth._update_current_tokens(None, None)
        return str_response
    except boxsdk.exception.BoxOAuthException as e:
        # we've done the work, so let's wipe the temporary state adjustment clean
        app.oauth._update_current_tokens(None, None)
        # response.status = e.status
        raise HTTPException(status_code=e.status, detail=json.dumps(e.message))


conf_obj = configparser.ConfigParser()


def main():
    global conf_obj

    conf_dir = Path("~/.config/diycrate_server").expanduser().resolve()
    if not conf_dir.is_dir():
        conf_dir.mkdir()
    cloud_credentials_file_path = conf_dir / "box.ini"
    if not cloud_credentials_file_path.is_file():
        cloud_credentials_file_path.write_text("")
    conf_obj.read(cloud_credentials_file_path)
    args = get_arg_parser()

    had_oauth2 = conf_obj.has_section("oauth2")
    if not had_oauth2:
        conf_obj.add_section("oauth2")
    conf_obj["oauth2"] = {
        "client_id": args.client_id or conf_obj["oauth2"]["client_id"],
        "client_secret": args.client_secret or conf_obj["oauth2"]["client_secret"],
    }
    configure_ssl_conf(args, conf_obj)

    with cloud_credentials_file_path.open("w") as fh:
        conf_obj.write(fh)
    uvicorn.run(
        app=app,
        port=args.port,
        host="0.0.0.0",
        ssl_certfile=conf_obj["ssl"]["cacert_pem_path"],
        ssl_keyfile=conf_obj["ssl"]["privkey_pem_path"],
        ssl_ca_certs=conf_obj["ssl"].get("chain_pem_path"),
    )


def configure_ssl_conf(args, conf_obj):
    try:
        prev_chain_pem_path = conf_obj["ssl"]["chain_pem_path"]
    except KeyError:
        prev_chain_pem_path = None

    if "ssl" not in conf_obj:
        if not args.cacert_pem_path:
            raise ValueError("Need a valid cacert_pem_path")
        if not args.privkey_pem_path:
            raise ValueError("Need a valid privkey_pem_path")
        conf_obj["ssl"] = {
            "cacert_pem_path": Path(args.cacert_pem_path).expanduser().absolute(),
            "privkey_pem_path": Path(args.privkey_pem_path).expanduser().absolute(),
        }
        if args.chain_pem_path:
            conf_obj["ssl"]["chain_pem_path"] = (
                Path(args.chain_pem_path).expanduser().resolve()
            )

    conf_obj["ssl"] = {
        "cacert_pem_path": Path(args.cacert_pem_path).expanduser().resolve()
        if args.cacert_pem_path
        else conf_obj["ssl"]["cacert_pem_path"],
        "privkey_pem_path": Path(args.privkey_pem_path).expanduser().resolve()
        if args.privkey_pem_path
        else conf_obj["ssl"]["privkey_pem_path"],
    }

    if args.chain_pem_path or prev_chain_pem_path:
        conf_obj["ssl"]["chain_pem_path"] = (
            prev_chain_pem_path or Path(args.chain_pem_path).expanduser().resolve()
        )


def get_arg_parser():
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument(
        "--client_id",
        type=str,
        help="Client ID provided by {}".format(cloud_provider_name),
        default="",
    )
    arg_parser.add_argument(
        "--client_secret",
        type=str,
        help="Client Secret provided by {}".format(cloud_provider_name),
        default="",
    )
    arg_parser.add_argument(
        "--cacert_pem_path",
        type=str,
        help="filepath to where the cacert.pem is located",
        default="",
    )
    arg_parser.add_argument(
        "--privkey_pem_path",
        type=str,
        help="filepath to where the privkey.pem is located",
        default="",
    )
    arg_parser.add_argument(
        "--chain_pem_path",
        type=str,
        help="filepath to where the chain.pem is located",
        default="",
    )
    arg_parser.add_argument("--port", type=int, help="server port", default=8081)
    args = arg_parser.parse_args()
    return args


if __name__ == "__main__":
    main()



import configparser
import json
import logging
import os
import shutil
import time
from datetime import datetime, timedelta
from functools import partial
from pathlib import Path
from typing import Generator, Union, Mapping

import boxsdk
from boxsdk.object.file import File
from boxsdk.object.folder import Folder
from boxsdk import Client, exception
from boxsdk.object.event import Event
from boxsdk.object.events import UserEventsStreamType, Events
from boxsdk.session.box_response import BoxResponse
from boxsdk.util.api_call_decorator import api_call
from boxsdk.util.lru_cache import LRUCache
from dateutil.parser import parse
from dateutil.tz import tzutc
from requests import Timeout
from send2trash import send2trash

from diycrate.cache_utils import (
    redis_set,
    r_c,
    redis_key,
    redis_get,
    local_or_box_file_m_time_key_func,
    redis_path_for_object_id_key,
)
from diycrate.file_operations import EventHandler
from diycrate.gui import notify_user_with_gui
from diycrate.item_queue_io import (
    DownloadQueueItem,
    download_pool_executor,
    download_queue_processor,
)
from diycrate.log_utils import setup_logger
from diycrate.oauth_utils import get_access_token
from diycrate.utils import FastAPI

setup_logger()

crate_logger = logging.getLogger(__name__)

conf_obj = configparser.ConfigParser()
conf_dir = Path("~/.config/diycrate").expanduser().resolve()
if not conf_dir.is_dir():
    conf_dir.mkdir()
cloud_credentials_file_path = conf_dir / "box.ini"
if not cloud_credentials_file_path.is_file():
    cloud_credentials_file_path.write_text("")
conf_obj.read(cloud_credentials_file_path)
BOX_DIR = Path(conf_obj["box"]["directory"]).expanduser().resolve()


def process_long_poll_event(client: Client, event: Union[Event, Mapping]):
    if event["event_type"] == "ITEM_CREATE":
        process_item_create_long_poll(client, event)
    if event["event_type"] == "ITEM_COPY":
        process_item_copy_long_poll(client, event)
    if event["event_type"] == "ITEM_MOVE":
        process_item_move_long_poll(event)
    if event["event_type"] == "ITEM_RENAME":
        process_item_rename_long_poll(client, event)
    elif event["event_type"] == "ITEM_UPLOAD":
        process_item_upload_long_poll(client, event)
    elif event["event_type"] == "ITEM_TRASH":
        process_item_trash_long_poll(event)
    elif event["event_type"] == "ITEM_DOWNLOAD":
        pass


def process_item_create_long_poll(client: Client, event: Union[Event, Mapping]):
    obj_id = event["source"]["id"]
    obj_type = event["source"]["type"]
    if r_c.exists(f"diy_crate:event_ids:{event.event_id}"):
        return

    if int(event["source"]["path_collection"]["total_count"]) > 1:
        path = Path(
            *list(
                folder["name"]
                for folder in event["source"]["path_collection"]["entries"][1:]
            )
        )
    else:
        path = Path()
    path = BOX_DIR / path
    if not path.is_dir():
        os.makedirs(path)
    file_path = path / event["source"]["name"]
    if obj_type == "folder":
        if not file_path.is_dir():
            os.makedirs(file_path)
        try:
            box_item = client.folder(folder_id=obj_id).get(
                fields=["id", "modified_at", "etag", "name"]
            )
            redis_set(
                r_c,
                box_item,
                datetime.fromtimestamp(os.path.getmtime(file_path))
                .astimezone(tzutc())
                .timestamp(),
                BOX_DIR,
                True,
                path,
            )
            time_data_map = {
                Path(file_path)
                .resolve()
                .as_posix(): datetime.fromtimestamp(os.path.getmtime(path))
                .astimezone(tzutc())
                .timestamp()
            }
            for mkey, mvalue in time_data_map.items():
                r_c.set(local_or_box_file_m_time_key_func(mkey, False), mvalue)
            r_c.set(
                local_or_box_file_m_time_key_func(path / box_item.name, True),
                parse(box_item.modified_at).astimezone(tzutc()).timestamp(),
            )
            r_c.set(
                redis_path_for_object_id_key(path / box_item.name), box_item.object_id
            )
            r_c.setex(
                f"diy_crate:event_ids:{event.event_id}",
                timedelta(days=32),
                path.as_posix(),
            )
        except boxsdk.exception.BoxAPIException as box_exc:
            if box_exc.status == 404 and box_exc.code == "trashed":
                crate_logger.debug(
                    f"Object {obj_id=} was previously deleted from Box. "
                    f"{path / event['source']['name']=}"
                )
                r_c.delete(redis_key(obj_id))
                r_c.set("diy_crate.last_save_time_stamp", int(time.time()))
                r_c.setex(
                    f"diy_crate:event_ids:{event.event_id}",
                    timedelta(days=32),
                    path.as_posix(),
                )
            else:
                raise

    elif obj_type == "file":
        if not file_path.is_file():
            try:
                queue_item = DownloadQueueItem(
                    client.file(file_id=obj_id).get(),
                    path / event["source"]["name"],
                    client._oauth,
                    event,
                )
                download_pool_executor.apply_async(
                    download_queue_processor, kwds=dict(queue_item=queue_item)
                )
            except boxsdk.exception.BoxAPIException as box_exc:
                if box_exc.status == 404 and box_exc.code == "trashed":
                    crate_logger.debug(
                        f"Object {obj_id=} was previously deleted from Box. "
                        f"{path / event['source']['name']=}"
                    )
                    r_c.delete(redis_key(obj_id))
                    r_c.set("diy_crate.last_save_time_stamp", int(time.time()))
                    r_c.setex(
                        f"diy_crate:event_ids:{event.event_id}",
                        timedelta(days=32),
                        path.as_posix(),
                    )
                else:
                    raise


def process_item_copy_long_poll(client: Client, event: Union[Event, Mapping]):
    obj_id = event["source"]["id"]
    obj_type = event["source"]["type"]
    if r_c.exists(f"diy_crate:event_ids:{event.event_id}"):
        return

    if int(event["source"]["path_collection"]["total_count"]) > 1:
        path = Path(
            *list(
                folder["name"]
                for folder in event["source"]["path_collection"]["entries"][1:]
            )
        )
    else:
        path = Path()
    path = BOX_DIR / path
    if not path.is_dir():
        os.makedirs(path)
    file_path = path / event["source"]["name"]
    if obj_type == "folder":
        if not file_path.is_dir():
            os.makedirs(file_path)
        try:
            box_item = client.folder(folder_id=obj_id).get(
                fields=["id", "modified_at", "etag", "name"]
            )
            redis_set(
                r_c,
                box_item,
                datetime.fromtimestamp(os.path.getmtime(file_path))
                .astimezone(tzutc())
                .timestamp(),
                BOX_DIR,
                True,
                path,
            )
            time_data_map = {
                Path(file_path)
                .resolve()
                .as_posix(): datetime.fromtimestamp(os.path.getmtime(path))
                .astimezone(tzutc())
                .timestamp()
            }
            for mkey, mvalue in time_data_map.items():
                r_c.set(local_or_box_file_m_time_key_func(mkey, False), mvalue)
            r_c.set(
                local_or_box_file_m_time_key_func(path / box_item.name, True),
                parse(box_item.modified_at).astimezone(tzutc()).timestamp(),
            )
            r_c.set(
                redis_path_for_object_id_key(path / box_item.name), box_item.object_id
            )
            r_c.setex(
                f"diy_crate:event_ids:{event.event_id}",
                timedelta(days=32),
                path.as_posix(),
            )
        except boxsdk.exception.BoxAPIException as box_exc:
            if box_exc.status == 404 and box_exc.code == "trashed":
                crate_logger.debug(
                    f"Object {obj_id=} was previously deleted from Box. "
                    f"{path / event['source']['name']=}"
                )
                r_c.delete(redis_key(obj_id))
                r_c.set("diy_crate.last_save_time_stamp", int(time.time()))
                r_c.setex(
                    f"diy_crate:event_ids:{event.event_id}",
                    timedelta(days=32),
                    path.as_posix(),
                )
            else:
                raise

    elif obj_type == "file":
        if not file_path.is_file():
            try:
                box_item = client.file(file_id=obj_id).get()
                queue_item = DownloadQueueItem(
                    box_item,
                    path / event["source"]["name"],
                    client._oauth,
                    event,
                )
                download_pool_executor.apply_async(
                    download_queue_processor, kwds=dict(queue_item=queue_item)
                )

            except boxsdk.exception.BoxAPIException as box_exc:
                if box_exc.status == 404 and box_exc.code == "trashed":
                    crate_logger.debug(
                        f"Object {obj_id=} was previously deleted from Box. "
                        f"{path / event['source']['name']=}"
                    )
                    r_c.delete(redis_key(obj_id))
                    r_c.set("diy_crate.last_save_time_stamp", int(time.time()))
                    r_c.setex(
                        f"diy_crate:event_ids:{event.event_id}",
                        timedelta(days=32),
                        path.as_posix(),
                    )
                else:
                    raise


def process_item_trash_long_poll(event: Union[Event, Mapping]):
    obj_id = event["source"]["id"]
    obj_type = event["source"]["type"]
    if r_c.exists(f"diy_crate:event_ids:{event.event_id}"):
        return

    if obj_type == "file":
        process_item_trash_file(event, obj_id)
    elif obj_type == "folder":
        process_item_trash_folder(event, obj_id)


def process_item_trash_file(event: Union[Event, Mapping], obj_id):
    item_info = r_c.get(redis_key(obj_id))
    if item_info:
        item_info = json.loads(str(item_info, encoding="utf-8", errors="strict"))
    if int(event["source"]["path_collection"]["total_count"]) > 1:
        path = Path(
            *list(
                folder["name"]
                for folder in event["source"]["path_collection"]["entries"][1:]
            )
        )
    else:
        path = Path()
    path = BOX_DIR / path
    file_path: Path
    if not item_info:
        file_path = path / event["source"]["name"]
    else:
        file_path = path / item_info["file_path"]
    if file_path.exists():
        send2trash(file_path.as_posix())
    if r_c.exists(redis_key(obj_id)):
        r_c.delete(redis_key(obj_id), redis_path_for_object_id_key(file_path))
        r_c.set("diy_crate.last_save_time_stamp", int(time.time()))
    notify_user_with_gui(
        "Box message: Deleted:",
        file_path.as_posix(),
        crate_logger=crate_logger,
        expire_time=10000,
    )
    r_c.setex(
        f"diy_crate:event_ids:{event.event_id}", timedelta(days=32), path.as_posix()
    )


def process_item_trash_folder(event: Union[Event, Mapping], obj_id):
    item_info = r_c.get(redis_key(obj_id))
    if item_info:
        item_info = json.loads(str(item_info, encoding="utf-8", errors="strict"))
    if int(event["source"]["path_collection"]["total_count"]) > 1:
        path = Path(
            *list(
                folder["name"]
                for folder in event["source"]["path_collection"]["entries"][1:]
            )
        )
    else:
        path = Path()
    path = BOX_DIR / path
    if not item_info:
        file_path = path / event["source"]["name"]
    else:
        file_path = path / item_info["file_path"]
    if file_path.is_dir():
        # still need to get the parent_id
        for sub_box_id in get_sub_ids(obj_id):
            cur_sub_box_redis_data = r_c.get(redis_key(sub_box_id))
            if cur_sub_box_redis_data:
                cur_sub_box_item_info = json.loads(
                    str(cur_sub_box_redis_data, encoding="utf-8", errors="strict")
                )
                if cur_sub_box_item_info:
                    r_c.delete(cur_sub_box_item_info["file_path"])
            r_c.delete(redis_key(sub_box_id))
        r_c.delete(redis_key(obj_id))
        shutil.rmtree(file_path)
        obj_cache_data_raw = r_c.get(redis_key(obj_id))
        obj_cache_data = (
            json.loads(str(obj_cache_data_raw))
            if obj_cache_data_raw is not None
            else obj_cache_data_raw
        )
        parent_id = obj_cache_data.get("parent_id") if obj_cache_data else None
        if parent_id:
            parent_folder_raw = r_c.get(redis_key(parent_id))
            parent_folder = (
                json.loads(str(parent_folder_raw))
                if parent_folder_raw is not None
                else parent_folder_raw
            )
            sub_ids = (
                parent_folder.get("sub_ids", [])
                if parent_folder is not None
                else parent_folder
            )
            if sub_ids:
                sub_ids.remove(obj_id)
                r_c.set(redis_key(parent_id), json.dumps(parent_folder))
            r_c.set("diy_crate.last_save_time_stamp", int(time.time()))
        deletion_msg = ["Box message: Deleted:", file_path.as_posix()]
        crate_logger.info(" ".join(deletion_msg))
        notify_user_with_gui(*deletion_msg, crate_logger=crate_logger)
        r_c.setex(
            f"diy_crate:event_ids:{event.event_id}", timedelta(days=32), path.as_posix()
        )


def process_item_upload_long_poll(client: Client, event: Union[Event, Mapping]):
    event_source: Union[File, Folder] = event["source"]
    obj_id = event_source["id"]
    obj_type = event_source["type"]
    if r_c.exists(f"diy_crate:event_ids:{event.event_id}"):
        return

    if obj_type == "file":
        if int(event["source"]["path_collection"]["total_count"]) > 1:
            path = Path(
                *list(
                    folder["name"]
                    for folder in event["source"]["path_collection"]["entries"][1:]
                )
            )
        else:
            path = Path()
        path = BOX_DIR / path
        if not path.exists():  # just in case this is a file in a new subfolder
            os.makedirs(path)
        crate_logger.debug(
            f"Submitting {path / event['source']['name']=} onto the download queue."
        )
        try:
            box_file_for_download: File = event["source"]
            queue_item = DownloadQueueItem(
                box_file_for_download,
                path / event["source"]["name"],
                client._oauth,
                event,
            )

            download_pool_executor.apply_async(
                download_queue_processor, kwds=dict(queue_item=queue_item)
            )

        except boxsdk.exception.BoxAPIException as box_exc:
            if box_exc.status == 404 and box_exc.code == "trashed":
                crate_logger.debug(
                    f"Object {obj_id=} was previously deleted from Box. "
                    f"{path / event['source']['name']=}"
                )
                r_c.delete(redis_key(obj_id))
                r_c.set("diy_crate.last_save_time_stamp", int(time.time()))
                r_c.setex(
                    f"diy_crate:event_ids:{event.event_id}",
                    timedelta(days=32),
                    path.as_posix(),
                )
            else:
                raise
    elif obj_type == "folder":
        if int(event["source"]["path_collection"]["total_count"]) > 1:
            path = Path(
                *list(
                    folder["name"]
                    for folder in event["source"]["path_collection"]["entries"][1:]
                )
            )
        else:
            path = Path()
        path = BOX_DIR / path
        if not path.exists():  # just in case this is a file in a new subfolder
            os.makedirs(path)
        file_path = path / event["source"]["name"]
        box_message = "new version"
        if not file_path.exists():
            os.makedirs(file_path)
            try:
                box_item: Folder = client.folder(folder_id=obj_id).get(
                    fields=["id", "name", "etag", "modified_at"]
                )
                redis_set(
                    r_c,
                    box_item,
                    datetime.fromtimestamp(os.path.getmtime(file_path))
                    .astimezone(tzutc())
                    .timestamp(),
                    BOX_DIR,
                    True,
                    path,
                )
            except boxsdk.exception.BoxAPIException as box_exc:
                if box_exc.status == 404 and box_exc.code == "trashed":
                    crate_logger.debug(
                        f"Object {obj_id=} was previously deleted from Box. {file_path=}"
                    )
                    r_c.setex(
                        f"diy_crate:event_ids:{event.event_id}",
                        timedelta(days=32),
                        path.as_posix(),
                    )
                    r_c.delete(redis_key(obj_id))
                    r_c.set("diy_crate.last_save_time_stamp", int(time.time()))
                else:
                    raise
            box_message = "new folder"
        notify_user_with_gui(
            f"Box message: {box_message}:",
            file_path.as_posix(),
            crate_logger=crate_logger,
        )
        r_c.setex(
            f"diy_crate:event_ids:{event.event_id}", timedelta(days=32), path.as_posix()
        )


def process_item_rename_long_poll(client: Client, event: Union[Event, Mapping]):
    obj_id = event["source"]["id"]
    obj_type = event["source"]["type"]
    if r_c.exists(f"diy_crate:event_ids:{event.event_id}"):
        return
    if obj_type == "file":
        if int(event["source"]["path_collection"]["total_count"]) > 1:
            path = Path(
                *list(
                    folder["name"]
                    for folder in event["source"]["path_collection"]["entries"][1:]
                )
            )
        else:
            path = Path()
        path = BOX_DIR / path
        file_path = path / event["source"]["name"]
        try:
            file_obj = client.file(file_id=obj_id).get()
            src_file_path = (
                None
                if not r_c.exists(redis_key(obj_id))
                else Path(redis_get(r_c, file_obj)["file_path"])
            )
            if (
                src_file_path
                and src_file_path.exists()
                and (not file_path.exists() or not src_file_path.samefile(file_path))
            ):
                version_info = redis_get(r_c, obj=file_obj)
                os.rename(src_file_path, file_path)
                version_info["file_path"] = file_path.as_posix()
                version_info["etag"] = file_obj["etag"]
                r_c.set(redis_key(obj_id), json.dumps(version_info))
                r_c.set("diy_crate.last_save_time_stamp", int(time.time()))
                time_data_map = {
                    Path(file_path)
                    .resolve()
                    .as_posix(): datetime.fromtimestamp(os.path.getmtime(path))
                    .astimezone(tzutc())
                    .timestamp()
                }
                for mkey, mvalue in time_data_map.items():
                    r_c.set(local_or_box_file_m_time_key_func(mkey, False), mvalue)
                r_c.set(
                    local_or_box_file_m_time_key_func(path / file_obj.name, True),
                    parse(file_obj.modified_at).astimezone(tzutc()).timestamp(),
                )
                r_c.set(
                    redis_path_for_object_id_key(path / file_obj.name),
                    file_obj.object_id,
                )
                r_c.delete(
                    redis_path_for_object_id_key(src_file_path),
                    local_or_box_file_m_time_key_func(src_file_path, False),
                    local_or_box_file_m_time_key_func(src_file_path, True),
                )
                r_c.setex(
                    f"diy_crate:event_ids:{event.event_id}",
                    timedelta(days=32),
                    path.as_posix(),
                )

            else:
                try:
                    version_info = redis_get(r_c, obj=file_obj)
                except TypeError:
                    crate_logger.error(
                        f"Key likely did not exist in the cache for file_obj id "
                        f"{file_obj.object_id=}",
                        exc_info=True,
                    )
                else:
                    if (
                        not file_path.exists()
                        or file_obj["etag"] != version_info["etag"]
                    ):
                        crate_logger.info(
                            f"Downloading the version from box, "
                            f"on a rename operation, "
                            f"and the etag (local {version_info['etag']=}) "
                            f"(box {file_obj['etag']=}) is different"
                        )
                        queue_item = DownloadQueueItem(
                            file_obj, file_path, client._oauth, event
                        )
                        download_pool_executor.apply_async(
                            download_queue_processor, kwds=dict(queue_item=queue_item)
                        )

        except boxsdk.exception.BoxAPIException as box_exc:
            if box_exc.status == 404 and box_exc.code == "trashed":
                crate_logger.debug(
                    f"Object {obj_id=} was previously deleted from Box. {file_path=}"
                )
                r_c.setex(
                    f"diy_crate:event_ids:{event.event_id}",
                    timedelta(days=32),
                    path.as_posix(),
                )
                r_c.delete(redis_key(obj_id))
                r_c.set("diy_crate.last_save_time_stamp", int(time.time()))
            else:
                raise

    elif obj_type == "folder":
        if int(event["source"]["path_collection"]["total_count"]) > 1:
            path = Path(
                *list(
                    folder["name"]
                    for folder in event["source"]["path_collection"]["entries"][1:]
                )
            )
        else:
            path = Path()
        path = BOX_DIR / path
        file_path = path / event["source"]["name"]
        try:
            folder_obj = client.folder(folder_id=obj_id).get()
            src_file_path = (
                None
                if not r_c.exists(redis_key(obj_id))
                else Path(redis_get(r_c, folder_obj)["file_path"])
            )
            if src_file_path and src_file_path.exists():
                os.rename(src_file_path, file_path)
                version_info = redis_get(r_c, obj=folder_obj)
                version_info["file_path"] = file_path.as_posix()
                version_info["etag"] = folder_obj["etag"]
                r_c.set(redis_key(obj_id), json.dumps(version_info))
                r_c.set("diy_crate.last_save_time_stamp", int(time.time()))
                time_data_map = {
                    Path(file_path)
                    .resolve()
                    .as_posix(): datetime.fromtimestamp(os.path.getmtime(path))
                    .astimezone(tzutc())
                    .timestamp()
                }
                for mkey, mvalue in time_data_map.items():
                    r_c.set(local_or_box_file_m_time_key_func(mkey, False), mvalue)
                r_c.set(
                    local_or_box_file_m_time_key_func(path / folder_obj.name, True),
                    parse(folder_obj.modified_at).astimezone(tzutc()).timestamp(),
                )
                r_c.set(
                    redis_path_for_object_id_key(path / folder_obj.name),
                    folder_obj.object_id,
                )
                r_c.delete(local_or_box_file_m_time_key_func(src_file_path, False))
                r_c.delete(local_or_box_file_m_time_key_func(src_file_path, True))
                r_c.setex(
                    f"diy_crate:event_ids:{event.event_id}",
                    timedelta(days=32),
                    path.as_posix(),
                )
        except boxsdk.exception.BoxAPIException as box_exc:
            if box_exc.status == 404 and box_exc.code == "trashed":
                crate_logger.debug(
                    f"Object {obj_id=} was previously deleted from Box. {file_path=}"
                )
                r_c.setex(
                    f"diy_crate:event_ids:{event.event_id}",
                    timedelta(days=32),
                    path.as_posix(),
                )
                r_c.delete(redis_key(obj_id))
                r_c.set("diy_crate.last_save_time_stamp", int(time.time()))
            else:
                raise


def process_item_move_long_poll(event: Union[Event, Mapping]):
    obj_id = event["source"]["id"]
    obj_type = event["source"]["type"]
    if r_c.exists(f"diy_crate:event_ids:{event.event_id}"):
        return
    if obj_type in ("file", "folder"):
        if r_c.exists(redis_key(obj_id)):
            item_info = json.loads(
                str(r_c.get(redis_key(obj_id)), encoding="utf-8", errors="strict")
            )
            src_file_path = Path(
                json.loads(
                    str(r_c.get(redis_key(obj_id)), encoding="utf-8", errors="strict")
                )["file_path"]
            )
            if int(event["source"]["path_collection"]["total_count"]) > 1:
                path = Path(
                    *list(
                        folder["name"]
                        for folder in event["source"]["path_collection"]["entries"][1:]
                    )
                )
            else:
                path = Path()
            path = BOX_DIR / path
            file_path = path / event["source"]["name"]
            if (
                obj_type == "file"
                and not src_file_path.exists()
                and file_path.exists()
                and item_info["etag"] != event["source"]["etag"]
            ):
                item_info["file_path"] = file_path.as_posix()
                item_info["etag"] = event["source"]["etag"]
                crate_logger.info(f"{event['source']['etag']=} updated.")
                r_c.set(redis_key(obj_id), json.dumps(item_info))
                r_c.setex(
                    f"diy_crate:event_ids:{event.event_id}",
                    timedelta(days=32),
                    path.as_posix(),
                )
                time_data_map = {
                    Path(file_path)
                    .resolve()
                    .as_posix(): datetime.fromtimestamp(os.path.getmtime(path))
                    .astimezone(tzutc())
                    .timestamp()
                }
                for mkey, mvalue in time_data_map.items():
                    r_c.set(local_or_box_file_m_time_key_func(mkey, False), mvalue)
                r_c.set(
                    local_or_box_file_m_time_key_func(file_path, True),
                    parse(event["source"]["modified_at"])
                    .astimezone(tzutc())
                    .timestamp(),
                )
            elif src_file_path.exists():
                to_set = []
                if obj_type == "folder":
                    for sub_id in get_sub_ids(obj_id):
                        if r_c.exists(redis_key(sub_id)):
                            sub_item_info = json.loads(
                                str(
                                    r_c.get(redis_key(sub_id)),
                                    encoding="utf-8",
                                    errors="strict",
                                )
                            )
                            orig_len = len(src_file_path.as_posix().split(os.path.sep))
                            tail = os.path.sep.join(
                                sub_item_info["file_path"].split(os.path.sep)[orig_len:]
                            )
                            new_sub_path = file_path / tail
                            sub_item_info["file_path"] = new_sub_path.as_posix()
                            to_set.append(
                                partial(
                                    r_c.set,
                                    redis_key(sub_id),
                                    json.dumps(sub_item_info),
                                )
                            )
                            to_set.append(
                                partial(
                                    r_c.set,
                                    redis_path_for_object_id_key(new_sub_path),
                                    sub_id,
                                )
                            )
                shutil.move(src_file_path, file_path)
                for item in to_set:
                    item()
                item_info["file_path"] = file_path.as_posix()
                item_info["etag"] = event["source"]["etag"]
                r_c.set(redis_key(obj_id), json.dumps(item_info))
                time_data_map = {
                    Path(file_path)
                    .resolve()
                    .as_posix(): datetime.fromtimestamp(os.path.getmtime(path))
                    .astimezone(tzutc())
                    .timestamp()
                }
                for mkey, mvalue in time_data_map.items():
                    r_c.set(local_or_box_file_m_time_key_func(mkey, False), mvalue)
                r_c.set(
                    local_or_box_file_m_time_key_func(file_path, True),
                    parse(event["source"]["modified_at"])
                    .astimezone(tzutc())
                    .timestamp(),
                )
                r_c.set(redis_path_for_object_id_key(file_path), obj_id)
                r_c.delete(redis_path_for_object_id_key(src_file_path))
                r_c.delete(local_or_box_file_m_time_key_func(src_file_path, False))
                r_c.delete(local_or_box_file_m_time_key_func(src_file_path, True))
                r_c.setex(
                    f"diy_crate:event_ids:{event.event_id}",
                    timedelta(days=32),
                    path.as_posix(),
                )


def get_sub_ids(box_id):
    """
    Retrieve the box item ids that are stored in this sub path
    :param box_id:
    :return:
    """
    ids = []
    item_info = json.loads(
        str(r_c.get(redis_key(box_id)), encoding="utf-8", errors="strict")
    )
    sub_ids = item_info.get("sub_ids")
    sub_ids = sub_ids if sub_ids is not None else []
    for sub_id in sub_ids:
        sub_id_cache_raw = r_c.get(redis_key(sub_id))
        if sub_id_cache_raw is not None:
            sub_item_info = json.loads(
                str(sub_id_cache_raw, encoding="utf-8", errors="strict")
            )
            if os.path.isdir(sub_item_info["file_path"]):
                ids.extend(get_sub_ids(sub_id))
            ids.append(sub_id)
    return ids


BOX_EVENT_IGNORED_TYPES = ["ITEM_PREVIEW", "ITEM_DOWNLOAD"]


class CustomBoxEvents(Events):
    @api_call
    def generate_events_with_long_polling(
        self, stream_position=None, stream_type=UserEventsStreamType.ALL
    ):
        """
        Subscribe to events from the given stream position.

        :param stream_position:
            The location in the stream from which to start getting events.
            0 is the beginning of time. 'now' will
            return no events and just current stream position.
        :type stream_position:
            `unicode`
        :param stream_type:
            (optional) Which type of events to return.
            Defaults to `UserEventsStreamType.ALL`.

            NOTE: Currently, the Box API requires this to be one of the user
            events stream types. The request will fail if an enterprise events
            stream type is passed.
        :type stream_type:
            :enum:`UserEventsStreamType`
        :returns:
            Events corresponding to changes on Box in realtime, as they come in.
        :rtype:
            `generator` of :class:`Event`
        """
        event_ids = LRUCache()
        stream_position = (
            stream_position
            if stream_position is not None
            else self.get_latest_stream_position(stream_type=stream_type)
        )
        while True:
            options = self.get_long_poll_options(stream_type=stream_type)
            while True:
                try:
                    long_poll_response: BoxResponse = self.long_poll(  # noqa
                        options, stream_position
                    )
                except Timeout:
                    break
                else:
                    message = long_poll_response.json()["message"]
                    if message == "new_change":
                        next_stream_position = stream_position
                        for event, next_stream_position in self._get_all_events_since(
                            stream_position, stream_type=stream_type
                        ):
                            try:
                                event_ids.get(event["event_id"])
                            except KeyError:
                                yield event, next_stream_position
                                event_ids.set(event["event_id"])
                        stream_position = next_stream_position
                        break
                    elif message == "reconnect":
                        continue
                    else:
                        break


class CustomBoxClient(Client):
    def events(self):
        """
        Get an events object that can get the latest events from Box
        or set up a long polling event subscription.
        """
        return CustomBoxEvents(self._session)


def long_poll_event_listener(file_event_handler: EventHandler, app: FastAPI):
    """
    Receive and process remote cloud item events in real-time
    :return:
    """
    global oauth
    handler = file_event_handler
    client = CustomBoxClient(oauth=handler.oauth)
    long_poll_streamer = client.events()

    while True:
        try:
            crate_logger.info("Determining stream position.")
            next_streamed_position = r_c.get("diy_crate.box.next_stream_position")
            if next_streamed_position:
                stream_position = str(
                    next_streamed_position, encoding="utf-8", errors="strict"
                )
                crate_logger.debug(
                    f"Using cached latest stream position: {str(stream_position)=}"
                )

            else:
                crate_logger.info("About to get latest stream position.")
                stream_position = long_poll_streamer.get_latest_stream_position()  # type: ignore
                crate_logger.debug(
                    f"Using as latest stream position: {str(stream_position)=}"
                )
                r_c.setex(
                    "diy_crate.box.next_stream_position",
                    timedelta(days=32),
                    stream_position,
                )
            event_stream: Generator[
                Union[Mapping, Event], None, None
            ] = long_poll_streamer.generate_events_with_long_polling(stream_position)
            for event, next_stream_position in event_stream:
                if event.get("message", "").lower() == "reconnect":
                    break
                if event.event_type in BOX_EVENT_IGNORED_TYPES:
                    continue
                if r_c.exists(f"diy_crate:event_ids:{event.event_id}"):
                    continue
                ev_source = event["source"]
                ev_src_name = getattr(ev_source, "name", None)
                ev_src_etag = getattr(ev_source, "etag", None)
                ev_src_modified_at = getattr(ev_source, "modified_at", None)
                ev_src_id = getattr(ev_source, "id", None)
                event_message = (
                    f"{event=} happened! {event.event_type=} "
                    f"{event.created_at=}, {event.event_id=}, "
                    f"{ev_src_name=}, "
                    f"{ev_src_etag=}, "
                    f"{ev_src_modified_at=}, "
                    f"{ev_src_id=}"
                )
                crate_logger.debug(event_message)
                process_long_poll_event(client, event)
                crate_logger.debug(
                    f"Will set diy_crate.box.next_stream_position to {next_stream_position=}"
                )
                r_c.setex(
                    "diy_crate.box.next_stream_position",
                    timedelta(days=32),
                    str(next_stream_position),
                )
        except exception.BoxAPIException as e:
            crate_logger.warning("Box or AttributeError occurred.", exc_info=e)
            get_access_token(client.auth._access_token, app=app)
            client.auth._update_current_tokens(oauth.access_token, oauth._refresh_token)
        except (AttributeError):
            crate_logger.warning("Box or AttributeError occurred.", exc_info=True)
        except Exception:
            crate_logger.warning("General error occurred.", exc_info=True)



import json
import os
import re
from typing import Optional, Union
import boxsdk
import traceback

from urllib import parse


def auth_url(redirect_url: str) -> str:
    client_id = os.environ.get("CLIENT_ID")
    client_secret = os.environ.get("CLIENT_SECRET")
    if not client_id:
        raise ValueError(f"Bad {client_id=}")
    box_oauth = boxsdk.OAuth2(client_id=client_id, client_secret=client_secret)
    auth_url_resp = box_oauth.get_authorization_url(redirect_url=redirect_url)
    return json.dumps(auth_url_resp)


def authenticate_url(code: str) -> str:
    """

    :return:
    """
    client_id = os.environ.get("CLIENT_ID")
    client_secret = os.environ.get("CLIENT_SECRET")
    if not client_id:
        raise ValueError(f"Bad {client_id=}")
    box_oauth = boxsdk.OAuth2(client_id=client_id, client_secret=client_secret)
    auth_code = code
    return json.dumps(
        [
            el.decode(encoding="utf-8", errors="strict")
            if isinstance(el, bytes)
            else el
            for el in box_oauth.authenticate(auth_code=auth_code)
        ]
    )


def new_access(
    access_token: Optional[str],
    refresh_token: str,
) -> list[Union[None, str]]:
    """
    Performs refresh of tokens and returns the result
    :return:
    """
    client_id = os.environ.get("CLIENT_ID")
    client_secret = os.environ.get("CLIENT_SECRET")
    if not client_id:
        raise ValueError(f"Bad {client_id=}")
    box_oauth = boxsdk.OAuth2(client_id=client_id, client_secret=client_secret)
    access_token_to_refresh = access_token or ""
    box_oauth._update_current_tokens(str(access_token_to_refresh), str(refresh_token))
    try:
        refresh_response: tuple[Optional[str], Optional[str]] = box_oauth.refresh(
            access_token_to_refresh
        )
        str_response = [
            el.decode(encoding="utf-8", errors="strict")
            if isinstance(el, bytes)
            else el
            for el in refresh_response
        ]
        # we've done the work, so let's wipe the temporary state adjustment clean
        box_oauth._update_current_tokens(None, None)
        return str_response
    except boxsdk.exception.BoxOAuthException as e:
        raise e


def main(event, _):
    path = event["http"]["path"]
    if re.match(r"^/auth_url(/)?", path) and event["http"]["method"] == "GET":
        query_string = event["http"]["queryString"]
        redirect_url: str = parse.parse_qs(query_string)["redirect_url"][0]
        status_code = 200
        try:
            response = auth_url(redirect_url)
        except Exception:
            response = traceback.format_exc()
            status_code = 500
        return {
            "body": response,
            "statusCode": status_code,
            "headers": {"Content-Type": "application/json"},
        }
    elif re.match(r"^/authenticate(/)?", path) and event["http"]["method"] == "POST":
        body: str = event["http"]["body"]
        status_code = 200

        try:
            response = authenticate_url(json.loads(body)["code"])
        except Exception:
            response = traceback.format_exc()
            status_code = 500
        return {
            "body": response,
            "statusCode": status_code,
            "headers": {"Content-Type": "application/json"},
        }

    elif re.match(r"/new_access(/)?", path) and event["http"]["method"] == "POST":
        body: str = event["http"]["body"]
        status_code = 200
        try:
            body_json = json.loads(body)
            response = new_access(body_json["access_token"], body_json["refresh_token"])
            response = json.dumps(response)
        except boxsdk.exception.BoxOAuthException as e:
            status_code = e.status
            response = json.dumps(e.message)

        except Exception:
            status_code = 500
            response = traceback.format_exc()
        return {
            "body": response,
            "statusCode": status_code,
            "headers": {"Content-Type": "application/json"},
        }

    return {
        "body": "Unsupported operation",
        "statusCode": 400,
        "headers": {"Content-Type": "application/json"},
    }



import threading

from boxsdk import OAuth2
import fastapi


class FastAPI(fastapi.FastAPI):
    oauth: OAuth2
    csrf_token: str
    started_cloud_threads: bool = False
    processing_oauth_browser: bool = False
    processing_oauth_browser_lock: threading.Lock
    processing_oauth_refresh_lock: threading.Lock



import logging
import time
from typing import Union

from boxsdk import BoxAPIException
from boxsdk.object.file import File
from boxsdk.object.folder import Folder

from .log_utils import setup_logger

setup_logger()

crate_logger = logging.getLogger(__name__)


class SafeIter:
    def __init__(self, iterable, path=None, tries=5):
        self.iterable = iterable
        self.path = path
        self.tries = tries or 5

    def __iter__(self):
        return self

    def __next__(self) -> Union[File, Folder, None]:
        tries = self.tries
        buffered_exc = None
        for _ in range(tries):
            try:
                value = next(self.iterable)
            except BoxAPIException as e:
                buffered_exc = e
                crate_logger.warning(
                    "Box API exception in folder: {path}".format(path=self.path),
                    exc_info=True,
                )
                time.sleep(5)
            else:
                break
        else:
            crate_logger.warning(
                "Could not recover from Box API issues after {tries} "
                "tries on folder: {path}".format(tries=tries, path=self.path),
                exc_info=buffered_exc,
            )
            value = None
        return value



import configparser
import subprocess

from pathlib import Path
from typing import Union

from setuptools.command.install import install


class InstallWrapper(install):
    def run(self):
        super().run()
        certbot_runner()


def certbot_runner():
    conf_obj = configparser.ConfigParser()

    conf_dir = Path("~/.config/diycrate_server").expanduser().resolve()
    if not conf_dir.is_dir():
        conf_dir.mkdir()
    cloud_credentials_file_path = conf_dir / "box.ini"
    if not cloud_credentials_file_path.is_file():
        cloud_credentials_file_path.write_text("")
    conf_obj.read(cloud_credentials_file_path)

    domain = str(
        Path(
            conf_obj.get(
                "ssl",
                "cacert_pem_path",
                fallback="/etc/letsencrypt/live/diycrate.xyz/cert.pem",
            )
        ).parent.name
    )

    certificates_cmd_result: Union[subprocess.CompletedProcess, None]
    try:
        certificates_cmd_result = subprocess.run(
            ["certbot", "certificates"], check=True, capture_output=True
        )
    except subprocess.CalledProcessError as e:
        certificates_cmd_result = None
        print(
            f"certbot may not be installed, "
            f"or the command issued may not be configured correctly. "
            f":(. "
            f"return code: {e.returncode}"
        )
    if certificates_cmd_result and domain not in bytes.decode(
        certificates_cmd_result.stdout
    ):
        try:
            subprocess.run(["certbot", "--standalone"], check=True)
        except subprocess.CalledProcessError as e:
            print(
                f"certbot may not be installed, "
                f"or the command issued may not be configured correctly. "
                f":(. "
                f"return code: {e.returncode}"
            )
        else:
            print("certbot configuration complete :)")

    elif certificates_cmd_result:
        print("certbot already configured :)")



import os

LOG_LEVEL = os.environ.get("DIY_CRATE_LOGGING_LEVEL")
LOG_TO_CONSOLE = os.environ.get("DIY_CRATE_LOGGING_CONSOLE")

if not LOG_LEVEL:
    LOG_LEVEL = "INFO"

try:
    from systemd import journal
except ImportError:
    journal = None


LOGGING = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {
        "verbose": {
            "format": "%(levelname)s %(asctime)s %(module)s:%(lineno)d "
            "%(process)d %(thread)d %(message)s"
        },
        "simple": {"format": "%(levelname)s %(message)s"},
    },
    "handlers": {
        "console": {
            "level": LOG_LEVEL,
            "class": "logging.StreamHandler",
            "formatter": "verbose",
        }
    },
    "loggers": {
        "diycrate": {
            "handlers": ["console"],
            "propagate": True,
            "level": LOG_LEVEL,
        },
        "diycrate_app": {
            "handlers": ["console"],
            "propagate": True,
            "level": LOG_LEVEL,
        },
        "__main__": {
            "handlers": ["console"],
            "propagate": True,
            "level": LOG_LEVEL,
        },
    },
}
#
if journal:
    LOGGING["handlers"]["journal"] = {
        "level": LOG_LEVEL,
        "class": "systemd.journal.JournalHandler",
        "formatter": "verbose",
    }

    LOGGING["loggers"]["root"] = {
        "handlers": ["journal"],
    }
    if not LOG_TO_CONSOLE:
        LOGGING["loggers"]["diycrate_app"]["handlers"] = []
        LOGGING["loggers"]["__main__"]["handlers"] = []
        LOGGING["loggers"]["diycrate"]["handlers"] = []



import argparse
import configparser
import logging
import os
import threading
import time
from pathlib import Path
from typing import Union
import typing

import fastapi
import uvicorn
from boxsdk.object.folder import Folder
import httpx
import pyinotify
from boxsdk import BoxAPIException, Client, exception


from diycrate.cache_utils import r_c
from diycrate.file_operations import (
    EventHandler,
    wm,
    in_moved_to,
    in_moved_from,
    mask,
)

from diycrate.long_poll_processing import long_poll_event_listener
from diycrate.oauth_utils import (
    get_access_token,
    store_tokens_callback,
    setup_remote_oauth,
    oauth_dance,
)
from diycrate.oauth_utils import oauth  # noqa: F401
from diycrate.path_utils import re_walk

from diycrate.log_utils import setup_logger
from diycrate.utils import FastAPI

setup_logger()

crate_logger = logging.getLogger(__name__)

cloud_provider_name = "Box"


# The watch manager stores the watches and provides operations on watches

# keep the lint-ing & introspection from complaining
# that these attributes don't exist before run-time.

BOX_DIR = Path()

trash_directory = Path("~/.local/share/Trash/files").expanduser()
wait_time = os.environ.get("DIY_CRATE_FILE_IO_OPERATIONS_WAIT_TIME_MS")
if wait_time:
    wait_time = int(wait_time) / 1000
else:
    wait_time = 1


long_poll_thread = threading.Thread(target=long_poll_event_listener)
long_poll_thread.daemon = True

walk_thread = threading.Thread(target=re_walk)
walk_thread.daemon = True

app = FastAPI()
app.processing_oauth_browser_lock = threading.Lock()
app.processing_oauth_refresh_lock = threading.Lock()

handler = EventHandler(app=app, wait_time=wait_time)

file_notify_read_freq = 3

notifier = pyinotify.ThreadedNotifier(wm, handler, read_freq=file_notify_read_freq)
notifier.coalesce_events()


# only re-enable when wishing to test token refresh or access dance
# @app.get('/kill_tokens/', response_class=fastapi.responses.PlainTextResponse)
# def kill_tokens():
#     # will get all of the oauth instances, since they are all the same reference
#     app.oauth._update_current_tokens(None, None)
#     r_c.delete('diy_crate.auth.refresh_token', 'diy_crate.auth.access_token')
#     return 'OK'

conf_obj = configparser.ConfigParser()


@app.get("/index/", response_class=fastapi.responses.PlainTextResponse)
def index():
    """
    Good to have a simple end-point.
    :return:
    """
    return "Hello, World!"


@app.get("/", response_class=fastapi.responses.PlainTextResponse)
def oauth_handler(state: str, code: str):
    """
    RESTful end-point for the oauth handling
    :return:
    """
    assert app.csrf_token == state
    post_data = {"code": code}
    post_kwargs = {
        (
            "json" if bool(int(conf_obj["box"]["auth_data_as_json"])) else "data"
        ): post_data
    }
    auth_response = httpx.post(
        conf_obj["box"]["authenticate_url"],
        verify=True,
        timeout=10.0,
        **post_kwargs,
    )
    access_token, refresh_token = auth_response.json()
    store_tokens_callback(access_token, refresh_token)
    app.oauth._update_current_tokens(access_token, refresh_token)
    if not getattr(app, "started_cloud_threads", False):
        start_cloud_threads()
        app.started_cloud_threads = True
    app.processing_oauth_browser = False
    app.processing_oauth_browser_lock.release()
    return "OK"


def start_cloud_threads():
    """

    :param client_oauth:
    :return:
    """
    global oauth
    client = Client(oauth)
    handler.oauth = oauth

    app.oauth = oauth
    wm.add_watch(BOX_DIR.as_posix(), mask, rec=True, auto_add=True)
    client.auth._access_token = r_c.get("diy_crate.auth.access_token")
    client.auth._refresh_token = r_c.get("diy_crate.auth.refresh_token")
    if client.auth._access_token:
        client.auth._access_token = (
            client.auth._access_token.decode(encoding="utf-8")
            if isinstance(client.auth._access_token, bytes)
            else client.auth._access_token
        )
    if client.auth._refresh_token:
        client.auth._refresh_token = (
            client.auth._refresh_token.decode(encoding="utf-8")
            if isinstance(client.auth._refresh_token, bytes)
            else client.auth._refresh_token
        )
    failed = False
    box_folder: Union[Folder, None] = None
    loop_index = 0
    while not failed:
        try:
            box_folder_initial = typing.cast(Folder, client.folder(folder_id="0"))
            box_folder = typing.cast(Folder, box_folder_initial.get())  # type: ignore
        except BoxAPIException as e:
            crate_logger.info("Bad box api response.", exc_info=e)
            get_access_token(client.auth._access_token, app=app)
            client._auth = oauth
            time.sleep(min([pow(2, loop_index), 8]))
            loop_index += 1
        except Exception:
            crate_logger.warning(
                "Encountered error getting root box folder", exc_info=True
            )
            time.sleep(min([pow(2, loop_index), 8]))
            loop_index += 1
        else:
            failed = True
    # local trash can
    wm.add_watch(
        trash_directory.as_posix(),
        mask=in_moved_to | in_moved_from,
        rec=True,
        auto_add=True,
    )

    if not long_poll_thread.is_alive():  # start before doing anything else
        long_poll_thread._args = (handler, app)
        long_poll_thread.start()
    if not walk_thread.is_alive():
        walk_thread._args = (BOX_DIR, box_folder, oauth, app, handler)
        walk_thread.start()


def main():
    global oauth
    global BOX_DIR

    conf_dir = Path("~/.config/diycrate").expanduser().resolve()
    if not conf_dir.is_dir():
        conf_dir.mkdir()
    cloud_credentials_file_path = conf_dir / "box.ini"
    if not cloud_credentials_file_path.is_file():
        cloud_credentials_file_path.write_text("")
    conf_obj.read(cloud_credentials_file_path)
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument(
        "--box_dir", type=str, help="directory for the root of box", default="~/box"
    )
    arg_parser.add_argument(
        "--cacert_pem_path",
        type=str,
        help="filepath to where the cacert.pem is located",
        default="",
    )
    arg_parser.add_argument(
        "--privkey_pem_path",
        type=str,
        help="filepath to where the privkey.pem is located",
        default="",
    )
    arg_parser.add_argument(
        "--authenticate_url",
        type=str,
        help="Full URL to perform remote Oauth2 authentication",
        default="",
    )
    arg_parser.add_argument(
        "--token_url",
        type=str,
        help="Full URL to do remote Oauth2 token refresh/retrieve",
        default="",
    )
    arg_parser.add_argument(
        "--authorization_url",
        type=str,
        help="Full URL to retrieve the " "Oauth2-generated authorization url",
        default="",
    )
    arg_parser.add_argument("--port", type=int, help="local web server port")
    arg_parser.add_argument(
        "--web-server-port",
        type=int,
        help="remote web server port",
        required=False,
        default=8080,
    )
    arg_parser.add_argument(
        "--auth-data-as-json",
        type=int,
        help="send auth data as json or not to remove server",
        required=False,
    )
    args = arg_parser.parse_args()
    had_oauth2 = conf_obj.has_section("oauth2")
    if had_oauth2:  # this was likely before we used the RemoteOauth2 workflow,
        # so we don't want this info hanging around
        # but if there is only the client_id defined, leave it alone
        if conf_obj["oauth2"].keys() - {"client_id"}:
            conf_obj.remove_section("oauth2")
    if "ssl" not in conf_obj:
        if not args.cacert_pem_path:
            raise ValueError("Need a valid cacert_pem_path")
        if not args.privkey_pem_path:
            raise ValueError("Need a valid privkey_pem_path")
        conf_obj["ssl"] = {
            "cacert_pem_path": Path(args.cacert_pem_path)
            .expanduser()
            .resolve()
            .as_posix(),
            "privkey_pem_path": Path(args.privkey_pem_path)
            .expanduser()
            .resolve()
            .as_posix(),
        }
    conf_obj["ssl"] = {
        "cacert_pem_path": Path(args.cacert_pem_path).expanduser().resolve().as_posix()
        if args.cacert_pem_path
        else conf_obj["ssl"]["cacert_pem_path"],
        "privkey_pem_path": Path(args.privkey_pem_path)
        .expanduser()
        .resolve()
        .as_posix()
        if args.privkey_pem_path
        else conf_obj["ssl"]["privkey_pem_path"],
    }

    with open(cloud_credentials_file_path, "w") as fh:
        conf_obj.write(fh)
    had_box = conf_obj.has_section("box")
    if not had_box:
        conf_obj.add_section("box")

    conf_obj["box"] = {
        "directory": args.box_dir or conf_obj["box"]["directory"],
        "authenticate_url": args.authenticate_url
        or conf_obj["box"].get(
            "authenticate_url", "https://localhost:8081/authenticate"
        ),
        "authorization_url": args.authorization_url
        or conf_obj["box"].get("authorization_url", "https://localhost:8081/auth_url"),
        "token_url": args.token_url
        or conf_obj["box"].get("token_url", "https://localhost:8081/new_access"),
        "web_server_port": args.port or args.web_server_port,
        "auth_data_as_json": args.auth_data_as_json
        if args.auth_data_as_json is not None
        else conf_obj["box"]["auth_data_as_json"],
    }
    web_server_port: int = int(conf_obj["box"]["web_server_port"])
    web_server_thread = threading.Thread(
        target=uvicorn.run,
        kwargs=dict(
            app=app,
            port=web_server_port,
            host="0.0.0.0",
            ssl_certfile=conf_obj["ssl"]["cacert_pem_path"],
            ssl_keyfile=conf_obj["ssl"]["privkey_pem_path"],
        ),
    )
    web_server_thread.daemon = True
    with cloud_credentials_file_path.open("w") as fh:
        conf_obj.write(fh)
    BOX_DIR = Path(conf_obj["box"]["directory"]).expanduser().resolve()
    if not BOX_DIR.is_dir():
        BOX_DIR.mkdir()
    if web_server_thread and not web_server_thread.is_alive():
        web_server_thread.start()

    if not (
        r_c.exists("diy_crate.auth.access_token")
        and r_c.exists("diy_crate.auth.refresh_token")
    ):
        app.oauth = oauth
        oauth_dance(
            redis_client=r_c,
            conf=conf_obj,
            app=app,
            file_event_handler=handler,
        )
        oauth = app.oauth
    else:
        try:
            oauth = setup_remote_oauth(r_c, conf=conf_obj, app=app)
            app.oauth = oauth
            handler.oauth = oauth
            start_cloud_threads()
            app.started_cloud_threads = True
        except exception.BoxOAuthException:
            r_c.delete("diy_crate.auth.access_token", "diy_crate.auth.refresh_token")
            oauth_dance(
                redis_client=r_c,
                conf=conf_obj,
                app=app,
                file_event_handler=handler,
            )
    notifier.start()
    # notifier_thread = threading.Thread(target=notifier.loop)
    # notifier_thread.daemon = True
    # notifier_thread.start()

    while threading.active_count() > 1:
        time.sleep(0.3)
    notifier.stop()


if __name__ == "__main__":
    main()



import threading
import typing
import webbrowser
from configparser import ConfigParser
import sys
from typing import List, Union

import httpx

from boxsdk import OAuth2
from boxsdk.auth import RemoteOAuth2
from pyinotify import ProcessEvent
from redis import Redis

from diycrate.utils import FastAPI
from diycrate.cache_utils import r_c


import logging

from .log_utils import setup_logger

setup_logger()

crate_logger = logging.getLogger(__name__)


def setup_oauth(cache_client, conf_object, callback: typing.Callable[[str, str], None]):
    """
    sets up the oauth instance with credentials and runtime callback.
    :param cache_client:
    :param conf_object:
    :param callback:
    :return:
    """
    global oauth
    access_token = cache_client.get("diy_crate.auth.access_token")
    refresh_token = cache_client.get("diy_crate.auth.refresh_token")
    if isinstance(access_token, bytes):
        access_token = access_token.decode(encoding="utf-8")
    if isinstance(refresh_token, bytes):
        refresh_token = refresh_token.decode(encoding="utf-8")
    oauth = OAuth2(
        client_id=conf_object["oauth2"]["client_id"],
        client_secret=conf_object["oauth2"]["client_secret"],
        store_tokens=callback,
        access_token=access_token,
        refresh_token=refresh_token,
    )
    return oauth


def get_access_token(
    access_token: str | bytes | None,
    app: Union[None, FastAPI] = None,
) -> str:
    """

    :param access_token: token we will try to refresh/get anew
    :return:
    """
    from .file_operations import conf_obj

    global oauth

    if isinstance(access_token, bytes):
        access_token = access_token.decode(encoding="utf-8")

    conf_object = conf_obj
    remote_url = conf_object["box"]["token_url"]
    refresh_token_from_cache: Union[bytes, str, None] = r_c.get(
        "diy_crate.auth.refresh_token"
    )

    refresh_token: str
    if isinstance(refresh_token_from_cache, bytes):
        refresh_token = refresh_token_from_cache.decode(encoding="utf-8")
    elif isinstance(refresh_token_from_cache, str):
        refresh_token = typing.cast(str, refresh_token_from_cache)
    else:
        refresh_token = ""
    if app:
        acquired = app.processing_oauth_refresh_lock.acquire()
        if acquired:
            if (access_token, refresh_token) != (
                app.oauth.access_token,
                app.oauth._refresh_token,
            ):
                if app.oauth.access_token and app.oauth._refresh_token:
                    app.processing_oauth_refresh_lock.release()
                    crate_logger.info(
                        "Acquired oauth refresh lock, "
                        "and the tokens appear to have changed already, %s, %s, %s, %s"
                        "so releasing early, skipping refresh function.",
                        access_token,
                        app.oauth.access_token,
                        refresh_token,
                        app.oauth._refresh_token,
                    )
                    access_token_resolved = app.oauth.access_token
                    store_tokens_callback(
                        access_token_resolved, app.oauth._refresh_token
                    )
                    return access_token_resolved

    post_data = {"refresh_token": refresh_token, "access_token": access_token}
    post_kwargs = {
        (
            "json" if bool(int(conf_obj["box"]["auth_data_as_json"])) else "data"
        ): post_data
    }
    response = httpx.post(
        remote_url,
        verify=True,
        timeout=10.0,
        **post_kwargs,
    )
    try:
        response.raise_for_status()
        response_json: List[str] = response.json()
        access_token_resolved, refresh_token = response_json
        oauth_resolved = oauth
        if oauth_resolved:
            oauth_resolved._update_current_tokens(access_token_resolved, refresh_token)
            oauth._update_current_tokens(access_token_resolved, refresh_token)
            if app:
                app.oauth._update_current_tokens(access_token_resolved, refresh_token)
                store_tokens_callback(access_token_resolved, refresh_token)
                app.processing_oauth_refresh_lock.release()
        else:
            raise AttributeError(
                "oauth_resolved has no attribute `None` (by hand exception)"
            )
    except AttributeError as e:
        if app:
            app.processing_oauth_refresh_lock.release()
        crate_logger.warn("Bad oauth object passed in.", exc_info=e)
        if r_c.exists("diy_crate.auth.access_token") and r_c.exists(
            "diy_crate.auth.refresh_token"
        ):
            r_c.delete("diy_crate.auth.access_token", "diy_crate.auth.refresh_token")
        sys.exit(1)
    except (ValueError, httpx.HTTPStatusError) as e:
        if app:
            app.processing_oauth_refresh_lock.release()
        response_error_body = response.content.decode()
        crate_logger.warn(
            "Error on response: %s, refresh_token len: %d, access_token len: %d",
            response_error_body,
            len(refresh_token),
            len(access_token or ""),
            exc_info=e,
        )

        # if r_c.exists("diy_crate.auth.access_token") and r_c.exists(
        #     "diy_crate.auth.refresh_token"
        # ):
        #     r_c.delete("diy_crate.auth.access_token", "diy_crate.auth.refresh_token")
        if app:
            handler = None
            crate_logger.info("Trying the regular oauth_dance.")
            oauth_dance(
                redis_client=r_c,
                conf=conf_obj,
                app=app,
                file_event_handler=handler,
            )
            raise e

        else:
            crate_logger.warn("Bad app object passed in, exiting", exc_info=e)
            sys.exit(1)
    else:
        crate_logger.debug(
            f"new access_token: {access_token_resolved}, refresh_token: {refresh_token}"
        )

        return access_token_resolved


def setup_remote_oauth(
    cache_client,
    retrieve_access_token: typing.Callable = get_access_token,
    conf: Union[ConfigParser, dict, None] = None,
    app: Union[FastAPI, None] = None,
) -> RemoteOAuth2:
    """
    sets up the oauth instance with credentials and runtime callback.
    :param cache_client:
    :param retrieve_access_token:
    :param conf:
    :return:
    """
    global oauth
    conf_obj: Union[ConfigParser, dict] = conf or {}
    access_token = cache_client.get("diy_crate.auth.access_token")

    if isinstance(access_token, bytes):
        access_token = access_token.decode(encoding="utf-8")
    refresh_token = cache_client.get("diy_crate.auth.refresh_token")

    if isinstance(refresh_token, bytes):
        refresh_token = refresh_token.decode(encoding="utf-8")
    if isinstance(conf_obj, ConfigParser):
        client_id = (
            (conf_obj["oauth2"]["client_id"] or "")
            if conf_obj.has_section("oauth2")
            else ""
        )
    else:
        # raise Exception
        client_id = conf_obj.get("oauth2", {}).get("client_id", "")

    def wrapper_retrieve_access_token(access_token_arg):
        return retrieve_access_token(
            access_token_arg,
            app=app,
        )

    oauth = RemoteOAuth2(
        client_id=client_id,
        client_secret="",
        access_token=access_token,
        refresh_token=refresh_token,
        retrieve_access_token=wrapper_retrieve_access_token,
    )
    return oauth


def setup_remote_oauth_initial_basic(
    cache_client,
    retrieve_access_token: typing.Callable = get_access_token,
    conf: Union[ConfigParser, dict, None] = None,
    app: Union[FastAPI, None] = None,
) -> RemoteOAuth2:
    """
    sets up the oauth instance with credentials and runtime callback.
    :param cache_client:
    :param retrieve_access_token:
    :param conf:
    :return:
    """

    conf_obj: Union[ConfigParser, dict] = conf or {}
    access_token = cache_client.get("diy_crate.auth.access_token")

    if isinstance(access_token, bytes):
        access_token = access_token.decode(encoding="utf-8")

    refresh_token = cache_client.get("diy_crate.auth.refresh_token")

    if isinstance(refresh_token, bytes):
        refresh_token = refresh_token.decode(encoding="utf-8")

    if isinstance(conf_obj, ConfigParser):
        client_id = (
            (conf_obj["oauth2"]["client_id"] or "")
            if conf_obj.has_section("oauth2")
            else ""
        )
    else:
        # raise Exception
        client_id = conf_obj.get("oauth2", {}).get("client_id", "")

    def wrapper_retrieve_access_token_function(access_token_arg):
        crate_logger.info(
            "This shim temporary function in a hack function was called, fun."
        )
        return retrieve_access_token(
            access_token_arg,
            app=app,
        )

    oauth_obj = RemoteOAuth2(
        client_id=client_id,
        client_secret="",
        access_token=access_token,
        refresh_token=refresh_token,
        retrieve_access_token=wrapper_retrieve_access_token_function,
    )
    return oauth_obj


def store_tokens_callback(
    access_token: Union[str, bytes], refresh_token: Union[str, bytes]
):
    """
    Intention is to store the oauth tokens
    :param access_token:
    :param refresh_token:
    :return:
    """
    assert access_token
    assert refresh_token
    if isinstance(access_token, bytes):
        access_token = access_token.decode(encoding="utf-8")

    if isinstance(refresh_token, bytes):
        refresh_token = refresh_token.decode(encoding="utf-8")
    r_c.set("diy_crate.auth.access_token", access_token)
    r_c.set("diy_crate.auth.refresh_token", refresh_token)


def oauth_dance(
    redis_client: Redis,
    conf: Union[ConfigParser, dict, None],
    app: FastAPI,
    file_event_handler: Union[ProcessEvent, None],
    web_server_thread: Union[threading.Thread, None] = None,
):
    global oauth
    if True:
        access_token, refresh_token = (
            app.oauth.access_token,
            app.oauth._refresh_token,
        )
        acquired = app.processing_oauth_browser_lock.acquire()
        if acquired:
            if (access_token, refresh_token) != (
                app.oauth.access_token,
                app.oauth._refresh_token,
            ):
                crate_logger.info(
                    f"Acquired oauth browser lock, "
                    f"and the tokens appear to have changed already, {access_token=}, "
                    f"{app.oauth.access_token=}, {refresh_token=}, "
                    f"{app.oauth._refresh_token=} "
                    f"so releasing early, skipping browser new oauth."
                )
                if oauth:
                    oauth._update_current_tokens(
                        app.oauth.access_token, app.oauth._refresh_token
                    )

                app.processing_oauth_browser_lock.release()
                return

    app.oauth = setup_remote_oauth(redis_client, conf=conf, app=app)
    if file_event_handler:
        file_event_handler.oauth = app.oauth
    redirect_url_domain = "127.0.0.1"
    (auth_url, csrf_token) = app.oauth.get_authorization_url(
        redirect_url=f"https://{redirect_url_domain}:{conf['box']['web_server_port']}/"
    )
    app.csrf_token = csrf_token
    if web_server_thread and not web_server_thread.is_alive():
        web_server_thread.start()
    webbrowser.open_new_tab(auth_url)  # make it easy for the end-user to start auth


oauth = setup_remote_oauth_initial_basic(cache_client=r_c)



import configparser
import json
import os
import queue
import threading
import logging
import time
from datetime import timedelta
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import Union, List, Dict
import typing


import boxsdk.object.file
import dateutil
import pyinotify
from boxsdk import Client
from boxsdk.exception import BoxAPIException
from boxsdk.object.file import File
from boxsdk.object.folder import Folder
from boxsdk.util.chunked_uploader import ChunkedUploader
from dateutil import tz
from dateutil.parser import parse
from requests.exceptions import ConnectionError
from urllib3.exceptions import ProtocolError

from diycrate.utils import FastAPI

from .cache_utils import (
    redis_key,
    redis_get,
    r_c,
    local_or_box_file_m_time_key_func,
    redis_set,
    redis_path_for_object_id_key,
)
from .item_queue_io import (
    UploadQueueItem,
    UploadQueueItemReal,
    upload_pool_executor,
    upload_queue_processor,
)
from .iter_utils import SafeIter
from .log_utils import setup_logger
from .oauth_utils import get_access_token, setup_remote_oauth

setup_logger()

crate_logger = logging.getLogger(__name__)

trash_directory = Path("~/.local/share/Trash/files").expanduser().resolve()

conf_obj = configparser.ConfigParser()
conf_dir = Path("~/.config/diycrate").expanduser().resolve()
if not conf_dir.is_dir():
    conf_dir.mkdir()
cloud_credentials_file_path = conf_dir / "box.ini"
if not cloud_credentials_file_path.is_file():
    cloud_credentials_file_path.write_text("")
conf_obj.read(cloud_credentials_file_path)
BOX_DIR = Path(conf_obj["box"]["directory"]).expanduser().resolve()


# noinspection PyUnresolvedReferences
class EventHandler(pyinotify.ProcessEvent):
    """
    EventHandler to manage cloud storage synchronization.
    """

    # noinspection SpellCheckingInspection
    def my_init(self, **kargs):
        """
        Extends the super to add cloud storage state.
        :return:
        """
        our_keys = ["oauth", "app", "wait_time"]
        our_kargs = {k: kargs.get(k) for k in our_keys}
        for key in our_keys:
            kargs.pop(key, None)
        super().my_init(**kargs)
        kargs = our_kargs
        self.move_events = []
        self.files_from_box = []
        self.folders_from_box = []
        self.incoming_operations = queue.Queue()
        self.operations = []
        self.wait_time = kargs.get("wait_time", 1)
        self.oauth = kargs.get("oauth")
        self.operations_thread = threading.Thread(
            target=self.incoming_operation_coalesce
        )
        self.app = kargs.get("app")
        self.operations_thread.daemon = True
        self.operations_thread.start()

    def incoming_operation_coalesce(self):
        while True:
            item = self.incoming_operations.get()
            self.operations.append(item)
            self.incoming_operations.task_done()
            self.operation_coalesce()

    def operation_coalesce(self):
        """
        Coalesce and process the operations in a more timely-fashion.
        :return:
        """
        # while True:
        #     # time.sleep(self.wait_time)
        cur_num_operations = len(self.operations)
        if cur_num_operations:
            operations_to_perform = self.operations[
                :cur_num_operations
            ]  # keep a local copy for this loop-run
            # operations list could have changed since the previous two instructions
            # pycharm complained that I was re-assigning the instance
            # variable outside the __init__.
            self.operations.clear()
            self.operations.extend(
                self.operations[cur_num_operations:]
                if len(self.operations) > cur_num_operations
                else []
            )
            for operation in operations_to_perform:
                self.process_event(*operation)

    @staticmethod
    def get_folder(client: Client, folder_id: str) -> Folder:
        """

        :param client:
        :param folder_id:
        :return:
        """
        folder = None
        num_retry = 15
        for x in range(num_retry):
            try:
                folder = client.folder(folder_id=folder_id).get()
                break
            except (
                ConnectionError,
                BrokenPipeError,
                ProtocolError,
                ConnectionResetError,
                BoxAPIException,
            ):
                crate_logger.warning(
                    "Error getting box folder id: {}".format(folder_id), exc_info=True
                )
                if x >= num_retry - 1:
                    crate_logger.debug(
                        "Failed for the last time to get the folder: {}".format(
                            folder_id
                        )
                    )
        return folder

    @staticmethod
    def folders_to_traverse(event_path: Union[str, Path]) -> List[str]:
        """

        :param event_path:
        :return:
        """
        if isinstance(event_path, str):
            event_path = Path(event_path)
        folders_to_traverse = [
            folder
            for folder in event_path.relative_to(BOX_DIR).parts
            if folder and folder != "/"
        ]
        return folders_to_traverse

    @staticmethod
    def traverse_path(
        client: Client,
        event: pyinotify.Event,
        cur_box_folder: Folder,
        folders_to_traverse: List["str"],
    ):
        """

        :param cur_box_folder:
        :param client:
        :param event:
        :param folders_to_traverse:
        :return:
        """
        for folder in folders_to_traverse:
            did_find_folder = False
            limit = 100
            offset = 0
            folder_items_generator = cur_box_folder.get_items(
                offset=offset, limit=limit
            )
            while True:
                try:
                    for entry in folder_items_generator:
                        if folder == entry["name"] and entry["type"] == "folder":
                            did_find_folder = True
                            cur_box_folder = EventHandler.get_folder(
                                client, entry.object_id
                            )
                            break  # found this piece of the path, keep going
                except BoxAPIException:
                    time.sleep(5)
                else:
                    break
            if not did_find_folder:
                try:
                    cur_box_folder = cur_box_folder.create_subfolder(folder).get()
                    crate_logger.debug(f"Sub-folder creation: {event.pathname}")
                except BoxAPIException:
                    crate_logger.debug(
                        "Box exception as we try to create a sub-folder {folder}".format(
                            folder=folder
                        ),
                        exc_info=True,
                    )
        return cur_box_folder

    def process_event(self, event: pyinotify.Event, operation: str):
        """
        Wrapper to process the given event on the operation.
        :param event:
        :param operation:
        :return:
        """
        crate_logger.debug(f"{operation=}, {event=}")
        try:
            if operation == "delete":
                self.process_delete_event(event)
            elif operation == "move":
                self.process_move_event(event)
            elif operation == "create":
                self.process_create_event(event)
            elif operation == "modify":
                self.process_modify_event(event, operation)
            elif operation == "real_close":
                self.process_real_close_event(event)
        except Exception:
            crate_logger.warning(
                "Operation: {operation} on event: {event} encountered an error.".format(
                    operation=operation, event=event
                ),
                exc_info=True,
            )

    def process_real_close_event(self, event: pyinotify.Event):
        file_path = Path(event.pathname)
        crate_logger.debug(f"Real close...: {file_path.as_posix()}")
        folders_to_traverse = self.folders_to_traverse(file_path.parent.as_posix())
        client = Client(self.oauth)
        cur_box_folder = get_box_folder(client, "0", 5, app=self.app)
        # if we're modifying in root box dir, then we've already found the folder
        try:
            is_base = any(
                Path(evp).relative_to(BOX_DIR) for evp in [event.path, event.path[:-1]]
            )
        except ValueError:
            is_base = False
        cur_box_folder = self.traverse_path(
            client, event, cur_box_folder, folders_to_traverse
        )
        last_dir = Path(event.path).parent.name
        if not is_base:
            AssertionError(
                cur_box_folder["name"] == last_dir,
                cur_box_folder["name"] + "not equals " + last_dir,
            )
        did_find_the_file = os.path.isdir(file_path)  # true if we are a directory :)
        did_find_the_folder = os.path.isfile(
            file_path
        )  # true if we are a regular file :)
        is_file = file_path.is_file()
        is_dir = file_path.is_dir()
        limit = 100
        offset = 0
        safe_iter = SafeIter(
            cur_box_folder.get_items(offset=offset, limit=limit), path=event.pathname
        )
        for entry in safe_iter:
            if entry is None:
                continue
            did_find_the_file = (
                is_file and entry["type"] == "file" and entry["name"] == file_path.name
            )
            did_find_the_folder = (
                is_dir and entry["type"] == "folder" and entry["name"] == file_path.name
            )
            if did_find_the_file:
                break
        # not a box file/folder (though could have been copied from a local box item)
        if is_file and not did_find_the_file:
            last_modified_time = (
                datetime.fromtimestamp(file_path.stat().st_mtime)
                .astimezone(dateutil.tz.tzutc())
                .timestamp()
            )
            try:
                if file_path.stat().st_size < 20000000:
                    raise BoxAPIException(code="file_size_too_small", status=400)
                box_uploader: ChunkedUploader = cur_box_folder.get_chunked_uploader(
                    file_path.as_posix()
                )
                uploader_callable = partial(
                    box_uploader.start,
                )
            except BoxAPIException as e:
                if e.status == 400 and e.code == "file_size_too_small":
                    uploader_callable = partial(
                        cur_box_folder.upload,
                        file_path.as_posix(),
                        file_path.name,
                        preflight_check=True,
                        upload_using_accelerator=True,
                    )
                else:
                    raise
            queue_item = UploadQueueItemReal(
                last_modified_time,
                uploader_callable,
                self.oauth,
                file_path.as_posix(),
            )

            upload_pool_executor.apply_async(
                upload_queue_processor, kwds=dict(queue_item=queue_item)
            )
        elif is_dir and not did_find_the_folder:
            queue_item: UploadQueueItem = partial(
                cur_box_folder.create_subfolder, file_path.name
            )

            upload_pool_executor.apply_async(
                upload_queue_processor, kwds=dict(queue_item=queue_item)
            )

    def process_modify_event(self, event: pyinotify.Event, operation: str):
        file_path = Path(event.pathname)
        crate_logger.debug(f"{operation}...: {file_path.as_posix()}")
        if r_c.exists(f"{event.pathname}:created:enqueue"):
            crate_logger.info(
                f"Skipping {operation} logic because cache key "
                f"`created:enqueue` for {file_path.as_posix()}"
            )
            return
        try:
            r_c.set(
                local_or_box_file_m_time_key_func(file_path.as_posix(), False),
                datetime.fromtimestamp(file_path.stat().st_mtime)
                .astimezone(dateutil.tz.tzutc())
                .timestamp(),
            )
        except FileNotFoundError:
            pass
        folders_to_traverse = self.folders_to_traverse(file_path.parent.as_posix())
        client = Client(self.oauth)

        folder_id = "0"
        retry_limit = 5
        cur_box_folder = get_box_folder(client, folder_id, retry_limit, app=self.app)
        # if we're modifying in root box dir, then we've already found the folder
        try:
            is_base = any(
                Path(evp).relative_to(BOX_DIR) for evp in [event.path, event.path[:-1]]
            )
        except ValueError:
            is_base = False
        cur_box_folder = self.traverse_path(
            client, event, cur_box_folder, folders_to_traverse
        )
        last_dir = Path(event.path).parent.name
        if not is_base:
            AssertionError(
                cur_box_folder["name"] == last_dir,
                cur_box_folder["name"] + "not equals " + last_dir,
            )
        did_find_the_file = os.path.isdir(file_path)  # true if we are a directory :)
        did_find_the_folder = os.path.isfile(
            file_path
        )  # true if we are a regular file :)
        is_file = os.path.isfile(file_path)
        is_dir = os.path.isdir(file_path)
        limit = 100
        offset = 0

        for entry in cur_box_folder.get_items(offset=offset, limit=limit):
            did_find_the_file = (
                is_file and entry["type"] == "file" and entry["name"] == file_path.name
            )
            did_find_the_folder = (
                is_dir and entry["type"] == "folder" and entry["name"] == file_path.name
            )
            if did_find_the_file:
                last_modified_time = file_path.stat().st_mtime
                if entry.object_id not in self.files_from_box:
                    cur_file = client.file(file_id=entry.object_id).get()
                    can_update = True
                    was_versioned = r_c.exists(redis_key(cur_file.object_id))
                    try:
                        info = redis_get(r_c, cur_file) if was_versioned else None
                        info = (
                            info
                            if was_versioned
                            else {"fresh_download": True, "etag": "0", "time_stamp": 0}
                        )
                        item_version = info
                        if cur_file["etag"] == item_version["etag"] and (
                            (
                                item_version["fresh_download"]
                                and (item_version["time_stamp"] >= last_modified_time)
                            )
                            or (
                                not item_version["fresh_download"]
                                and item_version["time_stamp"] >= last_modified_time
                            )
                        ):
                            can_update = False
                        if can_update:
                            queue_item = UploadQueueItemReal(
                                datetime.fromtimestamp(last_modified_time)
                                .astimezone(dateutil.tz.tzutc())
                                .timestamp(),
                                partial(cur_file.update_contents, file_path.as_posix()),
                                self.oauth,
                                file_path.as_posix(),
                            )

                            upload_pool_executor.apply_async(
                                upload_queue_processor, kwds=dict(queue_item=queue_item)
                            )
                        else:
                            is_new_time_stamp = (
                                item_version["time_stamp"] >= last_modified_time
                            )
                            crate_logger.debug(
                                "Skipping the update because not "
                                "versioned: {not_versioned}, "
                                "fresh_download: {fresh_download}, "
                                "version time_stamp >= "
                                "new time stamp: {new_time_stamp}, "
                                "event pathname: {path_name}, "
                                "cur file id: {obj_id}".format(
                                    not_versioned=not was_versioned,
                                    fresh_download=item_version["fresh_download"],
                                    new_time_stamp=is_new_time_stamp,
                                    path_name=event.pathname,
                                    obj_id=cur_file.object_id,
                                )
                            )
                    except TypeError:
                        crate_logger.debug("Error occurred", exc_info=True)
                    except Exception:
                        crate_logger.debug("Error occurred", exc_info=True)

                else:
                    self.files_from_box.remove(
                        entry.object_id
                    )  # just wrote if, assuming create event didn't run
                break
            elif did_find_the_folder:
                if entry.object_id not in self.folders_from_box:
                    crate_logger.debug(
                        "Cannot create a subfolder when it already exists: {}".format(
                            event.pathname
                        )
                    )
                else:
                    self.folders_from_box.remove(
                        entry.object_id
                    )  # just wrote if, assuming create event didn't run
                break
        if is_file and not did_find_the_file:
            crate_logger.debug("Uploading contents...: {}".format(file_path.as_posix()))
            last_modified_time = Path(event.pathname).stat().st_mtime
            queue_item = UploadQueueItemReal(
                datetime.fromtimestamp(last_modified_time)
                .astimezone(dateutil.tz.tzutc())
                .timestamp(),
                partial(
                    cur_box_folder.upload,
                    file_path.as_posix(),
                    file_path.name,
                    preflight_check=True,
                    upload_using_accelerator=True,
                ),
                self.oauth,
                file_path.as_posix(),
            )

            upload_pool_executor.apply_async(
                upload_queue_processor, kwds=dict(queue_item=queue_item)
            )
        if is_dir and not did_find_the_folder:
            crate_logger.debug(
                "Creating a sub-folder...: {}".format(file_path.as_posix())
            )
            queue_item: UploadQueueItem = partial(
                cur_box_folder.create_subfolder, file_path.name
            )

            upload_pool_executor.apply_async(
                upload_queue_processor, kwds=dict(queue_item=queue_item)
            )
            # wm.add_watch(file_path.as_posix(), rec=True, mask=mask, auto_add=True)

    def process_create_event(self, event: pyinotify.Event):
        crate_logger.debug("Creating: {}".format(event.pathname))
        folders_to_traverse = self.folders_to_traverse(event.path)
        client = Client(self.oauth)
        failed = False
        box_folder: Union[Folder, None] = None
        loop_index = 0
        while not failed:
            try:
                box_folder = typing.cast(Folder, client.folder(folder_id="0").get())  # type: ignore
            except BoxAPIException as e:
                crate_logger.warning("Error getting box root folder.", exc_info=e)
                get_access_token(client.auth.access_token, app=self.app)
                client._auth = self.app.oauth
                time.sleep(min([pow(2, loop_index), 8]))
                loop_index += 1
            except Exception:
                crate_logger.warning("Error getting box root folder.", exc_info=True)
                time.sleep(min([pow(2, loop_index), 8]))
                loop_index += 1
            else:
                failed = True
        cur_box_folder = box_folder
        # if we're modifying in root box dir, then we've already found the folder
        try:
            is_base = any(
                Path(evp).relative_to(BOX_DIR) for evp in [event.path, event.path[:-1]]
            )
        except ValueError:
            is_base = False
        cur_box_folder = self.traverse_path(
            client, event, cur_box_folder, folders_to_traverse
        )
        last_dir = Path(event.path).parent.name
        if not is_base:
            assert cur_box_folder["name"] == last_dir
        did_find_the_file = os.path.isdir(
            event.pathname
        )  # true if we are a directory :)
        did_find_the_folder = os.path.isfile(
            event.pathname
        )  # true if we are a regular file :)
        is_file = os.path.isfile(event.pathname)
        is_dir = os.path.isdir(event.pathname)
        limit = 100
        offset = 0
        for entry in cur_box_folder.get_items(offset=offset, limit=limit):
            did_find_the_file = (
                is_file
                and entry["type"] == "file"
                and entry["name"] == os.path.basename(event.pathname)
            )
            did_find_the_folder = (
                is_dir
                and entry["type"] == "folder"
                and entry["name"] == os.path.basename(event.pathname)
            )
            if did_find_the_file:
                dwld_key = "diy_crate.breadcrumb.create_from_box.{path}".format(
                    path=event.pathname
                )
                file_created_from_download = r_c.exists(dwld_key)
                if file_created_from_download:
                    self.files_from_box.append(entry.object_id)
                    r_c.delete(dwld_key)
                if entry.object_id not in self.files_from_box:
                    # more accurately, was this created offline?
                    AssertionError(
                        False,
                        "We should not be able to create a "
                        "file that exists in box; should be a close/modify.",
                    )
                    crate_logger.debug("Update the file: {}".format(event.pathname))
                    a_file = client.file(file_id=entry["id"]).get()
                    # seem it is possible to get more than one create
                    # (without having a delete in between)
                    queue_item: UploadQueueItem = partial(
                        a_file.update_contents, event.pathname
                    )

                    upload_pool_executor.apply_async(
                        upload_queue_processor, kwds=dict(queue_item=queue_item)
                    )
                    # cur_box_folder.upload(event.pathname, event.name)
                else:
                    crate_logger.debug(
                        "Created from box: {fpath}, but create flag ran on "
                        "our fs too, so we are skipping the logic to upload.".format(
                            fpath=event.pathname
                        )
                    )
                    self.files_from_box.remove(entry["id"])  # just downloaded it
                break
            elif did_find_the_folder:
                # we are not going to re-create the folder, but we are
                # also not checking if the contents in this
                # local creation are different from the contents in box.
                if entry["id"] in self.folders_from_box:
                    self.folders_from_box.remove(entry["id"])  # just downloaded it
                break
        if is_file and not did_find_the_file:
            crate_logger.debug("Upload the file: {}".format(event.pathname))
            last_modified_time = os.path.getctime(event.pathname)
            r_c.set(
                f"{event.pathname}:created:enqueue", value=1, ex=timedelta(minutes=1)
            )
            queue_item = UploadQueueItemReal(
                datetime.fromtimestamp(last_modified_time)
                .astimezone(dateutil.tz.tzutc())
                .timestamp(),
                partial(
                    cur_box_folder.upload,
                    event.pathname,
                    os.path.basename(event.pathname),
                    preflight_check=True,
                    upload_using_accelerator=True,
                ),
                self.oauth,
                event.pathname,
            )

            upload_pool_executor.apply_async(
                upload_queue_processor, kwds=dict(queue_item=queue_item)
            )
        elif is_dir and not did_find_the_folder:
            crate_logger.debug("Upload the folder: {}".format(event.pathname))
            queue_item: UploadQueueItem = partial(
                cur_box_folder.create_subfolder,
                os.path.basename(event.pathname),
            )

            upload_pool_executor.apply_async(
                upload_queue_processor, kwds=dict(queue_item=queue_item)
            )
            # wm.add_watch(event.pathname, rec=True, mask=mask, auto_add=True)

    def process_move_event(self, event: pyinotify.Event):
        crate_logger.debug("Doing a move on: {}".format(event))
        src_event, dest_event = event
        folders_to_traverse = self.folders_to_traverse(dest_event.path)
        client = Client(self.oauth)
        failed = False
        box_folder = None
        while not failed:
            try:
                box_folder = client.folder(folder_id="0").get()
            except Exception:
                crate_logger.warning("Error getting root box folder.", exc_info=True)
                time.sleep(2)
            else:
                failed = True
        cur_box_folder = box_folder
        # if we're modifying in root box dir, then we've already found the folder
        cur_box_folder = self.traverse_path(
            client, dest_event, cur_box_folder, folders_to_traverse
        )
        src_folders_to_traverse = self.folders_to_traverse(src_event.path)
        src_box_folder = box_folder
        src_box_folder = self.traverse_path(
            client, src_event, src_box_folder, src_folders_to_traverse
        )
        is_rename = src_event.path == dest_event.path
        # is_a_directory = 'IN_ISDIR'.lower() in dest_event.maskname.lower()
        did_find_src_file = os.path.isdir(
            dest_event.pathname
        )  # true if we are a directory :)
        did_find_src_folder = os.path.isfile(
            dest_event.pathname
        )  # true if we are a regular file :)
        is_file = os.path.isfile(dest_event.pathname)
        is_dir = os.path.isdir(dest_event.pathname)
        move_from_remote = False
        limit = 100
        offset = 0
        folder_gen = src_box_folder.get_items(offset=offset, limit=limit)
        safe_iter = SafeIter(folder_gen, path=dest_event.name)
        for entry in safe_iter:
            if entry is None:
                continue
            did_find_src_file = (
                is_file and entry["name"] == src_event.name and entry["type"] == "file"
            )
            did_find_src_folder = (
                is_dir and entry["name"] == src_event.name and entry["type"] == "folder"
            )
            if did_find_src_file:
                src_file: File = client.file(file_id=entry["id"]).get()
                if is_rename:
                    crate_logger.debug(f"Processing as a rename {dest_event.pathname=}")
                    last_modified_time = (
                        datetime.fromtimestamp(
                            Path(dest_event.pathname).stat().st_mtime
                        )
                        .astimezone(dateutil.tz.tzutc())
                        .timestamp()
                    )
                    file_obj = client.file(file_id=src_file.object_id).get()
                    data_sent_to_cache = redis_set(
                        r_c, file_obj, last_modified_time, box_dir_path=BOX_DIR
                    )
                    version_info = data_sent_to_cache[redis_key(src_file.object_id)]
                    r_c.set(
                        local_or_box_file_m_time_key_func(dest_event.pathname, False),
                        datetime.fromtimestamp(
                            Path(dest_event.pathname).stat().st_mtime
                        )
                        .astimezone(dateutil.tz.tzutc())
                        .timestamp(),
                    )
                    src_file.rename(dest_event.name)
                    file_obj: File = client.file(file_id=src_file.object_id).get()
                    version_info["file_path"] = dest_event.pathname
                    version_info["etag"] = file_obj["etag"]
                    r_c.set(
                        redis_path_for_object_id_key(dest_event.pathname),
                        src_file.object_id,
                    )
                    r_c.delete(
                        redis_path_for_object_id_key(
                            Path(src_event.path) / src_file.name
                        )
                    )
                    r_c.set(redis_key(src_file.object_id), json.dumps(version_info))
                    r_c.set("diy_crate.last_save_time_stamp", int(time.time()))
                    path_builder = BOX_DIR
                    oauth = setup_remote_oauth(r_c, conf=conf_obj, app=self.app)
                    client = Client(oauth)
                    parent_folders: list[Folder] = (
                        client.file(file_obj.object_id)
                        .get(fields=["path_collection"])
                        .path_collection["entries"]
                    )
                    for updated_entry in parent_folders:
                        if updated_entry.id == "0":
                            continue
                        path_builder /= updated_entry.name
                        folder_entry: Folder = client.folder(updated_entry.id).get(
                            fields=["modified_at"]
                        )
                        if not r_c.exists(redis_path_for_object_id_key(path_builder)):
                            r_c.set(
                                redis_path_for_object_id_key(path_builder),
                                updated_entry.object_id,
                            )
                        r_c.set(
                            local_or_box_file_m_time_key_func(path_builder, True),
                            parse(folder_entry.modified_at)
                            .astimezone(dateutil.tz.tzutc())
                            .timestamp(),
                        )

                else:
                    crate_logger.debug(f"Processing as a move {dest_event.pathname=}.")
                    did_find_cur_file = os.path.isdir(
                        dest_event.pathname
                    )  # should check box instead
                    cur_offset = 0
                    cur_box_folder_items: list[
                        Union[File, Folder]
                    ] = cur_box_folder.get_items(offset=cur_offset, limit=limit)
                    for cur_entry in cur_box_folder_items:
                        matching_name = cur_entry["name"] == os.path.basename(
                            dest_event.pathname
                        )
                        did_find_cur_file = (
                            is_file and matching_name and isinstance(cur_entry, File)
                        )
                        did_find_cur_folder = (
                            is_dir and matching_name and isinstance(cur_entry, Folder)
                        )
                        if did_find_cur_file:
                            crate_logger.info(
                                f"Did find cur file (moving to {dest_event.pathname=})"
                            )
                            queue_item = UploadQueueItemReal(
                                datetime.fromtimestamp(
                                    Path(dest_event.pathname).stat().st_mtime
                                )
                                .astimezone(dateutil.tz.tzutc())
                                .timestamp(),
                                partial(cur_entry.update_contents, dest_event.pathname),
                                self.oauth,
                                dest_event.pathname,
                            )

                            upload_pool_executor.apply_async(
                                upload_queue_processor, kwds=dict(queue_item=queue_item)
                            )
                            queue_item: UploadQueueItem = partial(src_file.delete)

                            upload_pool_executor.apply_async(
                                upload_queue_processor, kwds=dict(queue_item=queue_item)
                            )
                            break
                        elif did_find_cur_folder:
                            crate_logger.debug(
                                "do not currently support moving a same name folder "
                                "into parent with"
                                "folder inside of the same name -- would may need "
                                "to update the contents"
                            )
                            break
                    if is_file and not did_find_cur_file and did_find_src_file:
                        crate_logger.debug(
                            f"Did not find cur file {dest_event.pathname=} "
                            f"so we call Box API's src_file.move to cur box folder."
                        )

                        def complex_move():
                            before_move_path_collection_entries = list(
                                src_file.path_collection["entries"]
                            )
                            src_file_move_ret_value = src_file.move(cur_box_folder)

                            # oauth = setup_remote_oauth(r_c, conf=conf_obj)
                            # client = Client(oauth)
                            def and_after():
                                complex_path_builder = BOX_DIR
                                for (
                                    complex_entry
                                ) in before_move_path_collection_entries:
                                    if complex_entry.id == "0":
                                        continue
                                    complex_path_builder /= complex_entry.name
                                    complex_folder_entry = client.folder(
                                        complex_entry.id
                                    ).get(fields=["modified_at"])
                                    r_c.set(
                                        local_or_box_file_m_time_key_func(
                                            complex_path_builder, True
                                        ),
                                        parse(complex_folder_entry.modified_at)
                                        .astimezone(dateutil.tz.tzutc())
                                        .timestamp(),
                                    )
                                    r_c.set(
                                        redis_path_for_object_id_key(
                                            complex_path_builder
                                        ),
                                        complex_folder_entry.object_id,
                                    )

                            return src_file_move_ret_value, partial(and_after)

                        queue_item = UploadQueueItemReal(
                            datetime.fromtimestamp(
                                Path(dest_event.pathname).stat().st_mtime
                            )
                            .astimezone(dateutil.tz.tzutc())
                            .timestamp(),
                            partial(complex_move),
                            self.oauth,
                            dest_event.pathname,
                        )

                        upload_pool_executor.apply_async(
                            upload_queue_processor, kwds=dict(queue_item=queue_item)
                        )
                        # src_file.move(cur_box_folder)
                        # do not yet support moving and renaming in one go
                        assert src_file["name"] == dest_event.name
            elif did_find_src_folder:
                src_folder = client.folder(folder_id=entry["id"]).get()
                if is_rename:
                    src_folder.rename(dest_event.name)
                else:
                    src_folder.move(cur_box_folder)
                    # do not yet support moving and renaming in one go
                    assert src_folder["name"] == dest_event.name
            elif entry["name"] == dest_event.name:
                if (
                    Path(src_event.path) == Path(dest_event.path)
                    and (
                        Path(src_event.name).stem == dest_event.name
                        or (
                            Path(src_event.name).suffix == ".tmp"
                            and Path(dest_event.name).suffix != ".tmp"
                        )
                    )
                    and src_event.name != dest_event.name
                ):
                    crate_logger.info(
                        f"Detected conceptual same directory temp file rename from "
                        f"{src_event.name=} to {dest_event.name=}"
                    )
                    cur_offset = 0
                    cur_entry: Union[boxsdk.object.file.File, boxsdk.object.file.Item]
                    for cur_entry in cur_box_folder.get_items(
                        offset=cur_offset, limit=limit
                    ):
                        matching_name = cur_entry["name"] == os.path.basename(
                            dest_event.pathname
                        )
                        if matching_name:
                            crate_logger.debug(f"{cur_entry=} {type(cur_entry)=}")

                        did_find_cur_file = (
                            is_file and matching_name and isinstance(cur_entry, File)
                        )
                        did_find_cur_folder = (
                            is_dir and matching_name and isinstance(cur_entry, Folder)
                        )
                        if did_find_cur_file:
                            queue_item = UploadQueueItemReal(
                                datetime.fromtimestamp(
                                    Path(dest_event.pathname).stat().st_mtime
                                )
                                .astimezone(dateutil.tz.tzutc())
                                .timestamp(),
                                partial(cur_entry.update_contents, dest_event.pathname),
                                self.oauth,
                                dest_event.pathname,
                            )

                            upload_pool_executor.apply_async(
                                upload_queue_processor, kwds=dict(queue_item=queue_item)
                            )
                            break
                        elif did_find_cur_folder:
                            crate_logger.debug(
                                "do not currently support moving a same name folder "
                                "into parent with"
                                "folder inside of the same name -- would may need "
                                "to update the contents"
                            )
                            break
                else:
                    move_from_remote = True
        if (
            not move_from_remote
        ):  # if it was moved from a different folder on remote, could be false still
            limit = 100
            offset = 0
            for entry in cur_box_folder.get_items(offset=offset, limit=limit):
                if entry["name"] == dest_event.name:
                    move_from_remote = True
                    break
            if not move_from_remote:
                if is_file and not did_find_src_file:
                    # src file [should] no longer exist[s].
                    # this file did not originate in box, too.
                    last_modified_time = Path(dest_event.pathname).stat().st_mtime
                    queue_item = UploadQueueItemReal(
                        datetime.fromtimestamp(last_modified_time)
                        .astimezone(dateutil.tz.tzutc())
                        .timestamp(),
                        partial(
                            cur_box_folder.upload,
                            dest_event.pathname,
                            os.path.basename(dest_event.pathname),
                            preflight_check=True,
                            upload_using_accelerator=True,
                        ),
                        self.oauth,
                        dest_event.pathname,
                    )

                    upload_pool_executor.apply_async(
                        upload_queue_processor, kwds=dict(queue_item=queue_item)
                    )
                elif is_dir and not did_find_src_folder:
                    queue_item: UploadQueueItem = partial(
                        cur_box_folder.create_subfolder,
                        os.path.basename(dest_event.pathname),
                    )

                    upload_pool_executor.apply_async(
                        upload_queue_processor, kwds=dict(queue_item=queue_item)
                    )
                    # wm.add_watch(dest_event.pathname, rec=True, mask=mask, auto_add=True)

    def process_delete_event(self, event: pyinotify.Event):
        crate_logger.debug(f"Doing a delete on {event=}")
        folders_to_traverse = self.folders_to_traverse(event.path)
        client = Client(self.oauth)
        failed = False
        box_folder = None
        loop_index = 0
        while not failed:
            try:
                box_folder = client.folder(folder_id="0").get()
            except Exception:
                crate_logger.warning("Error getting box root folder.", exc_info=True)
                time.sleep(min([pow(2, loop_index), 8]))
                loop_index += 1
            else:
                failed = True
        cur_box_folder = box_folder
        # if we're modifying in root box dir, then we've already found the folder
        try:
            is_base = any(
                Path(evp).relative_to(BOX_DIR) for evp in [event.path, event.path[:-1]]
            )
        except ValueError:
            is_base = False
        cur_box_folder = self.traverse_path(
            client, event, cur_box_folder, folders_to_traverse
        )
        last_dir = Path(event.path).parent.name
        if not is_base:
            AssertionError(
                cur_box_folder["name"] == last_dir,
                cur_box_folder["name"] + "not equals " + last_dir,
            )
        event_was_for_dir = "IN_ISDIR".lower() in event.maskname.lower()
        limit = 100
        offset = 0
        for entry in cur_box_folder.get_items(offset=offset, limit=limit):
            if (
                not event_was_for_dir
                and entry["type"] == "file"
                and entry["name"] == os.path.basename(event.pathname)
            ):
                if entry["id"] not in self.files_from_box:
                    cur_file = client.file(file_id=entry["id"]).get()
                    if (
                        cur_file.delete()
                    ):  # does not actually check correctly...unless not "ok" is false
                        # del version_info[cur_file['id']]
                        r_c.delete(redis_key(cur_file["id"]))
                else:
                    self.files_from_box.remove(
                        entry["id"]
                    )  # just wrote if, assuming create event didn't run
                break
            elif (
                event_was_for_dir
                and entry["type"] == "folder"
                and entry["name"] == os.path.basename(event.pathname)
            ):
                if entry["id"] not in self.folders_from_box:
                    self.get_folder(client, entry["id"]).delete()
                else:
                    self.folders_from_box.remove(
                        entry["id"]
                    )  # just wrote if, assuming create event didn't run
                break

    # noinspection PyPep8Naming
    def process_IN_CREATE(self, event: pyinotify.Event):
        """
        Overrides the super.
        :param event:
        :return:
        """
        if not os.path.basename(event.pathname).startswith(
            ".~lock"
        ):  # avoid propagating lock files
            self.incoming_operations.put([event, "create"])

    # noinspection PyPep8Naming
    def process_IN_DELETE(self, event: pyinotify.Event):
        """
        Overrides the super.
        :param event:
        :return:
        """
        crate_logger.debug(event)
        self.incoming_operations.put([event, "delete"])

    # noinspection PyPep8Naming
    def process_IN_MODIFY(self, event: pyinotify.Event):
        """
        Overrides the super.
        :param event:
        :return:
        """
        if not os.path.basename(event.pathname).startswith(
            ".~lock"
        ):  # avoid propagating lock files
            self.incoming_operations.put([event, "modify"])

    # noinspection PyPep8Naming
    def process_IN_MOVED_FROM(self, event: pyinotify.Event):
        """
        Overrides the super.
        :param event:
        :return:
        """
        crate_logger.debug(event)
        if not os.path.basename(event.pathname).startswith(
            ".~lock"
        ):  # avoid propagating lock files
            crate_logger.debug("Moved from: {}".format(event.pathname))
            self.move_events.append(event)

    # noinspection PyPep8Naming
    def process_IN_MOVED_TO(self, event: pyinotify.Event):
        """
        Overrides the super.
        :param event:
        :return:
        """
        crate_logger.debug(event)
        found_from = False
        to_trash = trash_directory == Path(event.pathname).parent
        to_box = BOX_DIR in Path(event.pathname).parents
        move_event_processed = None
        for move_event in self.move_events:
            was_moved_from = "in_moved_from" in move_event.maskname.lower()
            if (
                move_event.cookie == event.cookie  # noqa
                and was_moved_from
                and BOX_DIR in Path(move_event.pathname).parents
            ):
                found_from = True
                # only count deletes that come from within the box path --
                # though this should always be the case
                if to_trash:
                    self.incoming_operations.put([move_event, "delete"])
                else:
                    self.incoming_operations.put([[move_event, event], "move"])
                move_event_processed = move_event
                break
        if move_event_processed:
            try:
                self.move_events.remove(move_event_processed)
            except ValueError:
                crate_logger.warning(
                    f"Found a matching src moved from {move_event_processed=}, "
                    f"but move_events no longer contained when we tried to remove, hmm.",
                    exc_info=True,
                )
        if not found_from and (not to_trash and to_box):
            self.incoming_operations.put(
                [event, "modify"]
            )  # "close"/"modify" seems appropriate
            # allow moving from a ~.lock file...I guess that may be okay
            crate_logger.debug("Moved to: {}".format(event.pathname))

    # noinspection PyPep8Naming,PyMethodMayBeStatic
    def process_IN_DELETE_SELF(self, event: pyinotify.Event):
        crate_logger.debug(event)

    # noinspection PyPep8Naming,PyMethodMayBeStatic
    def process_ALL_EVENTS(self, event: pyinotify.Event):
        crate_logger.debug(event)

    # noinspection PyPep8Naming
    def process_IN_CLOSE(self, event: pyinotify.Event):
        """
        Overrides the super.
        :param event:
        :return:
        """
        if not os.path.basename(event.pathname).startswith(
            ".~lock"
        ):  # avoid propagating lock files
            crate_logger.debug("Had a close on: {}".format(event))
            self.incoming_operations.put([event, "real_close"])
        else:
            crate_logger.debug(f"Skipping close on: {event.pathname=} due to `.~lock`")


def get_box_folder(
    client: Client,
    folder_id: str,
    retry_limit: int,
    app: Union[FastAPI, None] = None,
) -> Folder:
    """

    :param client:
    :param cur_box_folder:
    :param folder_id:
    :param retry_limit:
    :return:
    """
    cur_box_folder: Folder
    for i in range(retry_limit):
        try:
            box_folder_skeleton: Folder = typing.cast(
                Folder, client.folder(folder_id=folder_id)
            )
            box_folder: Folder = typing.cast(Folder, box_folder_skeleton.get())
            cur_box_folder = box_folder
            break
        except BoxAPIException as e:
            get_access_token(client.auth._access_token, app=app)
            if app:
                client._auth = app.oauth
            if i == retry_limit - 1:
                crate_logger.info("Bad box api response.", exc_info=e)
                raise e
        except (
            ConnectionError,
            BrokenPipeError,
            ProtocolError,
            ConnectionResetError,
        ) as e:
            if i + 1 >= retry_limit:
                crate_logger.warning(
                    "Attempt ({retry_count}) out of ({max_count}); Going to give "
                    "up on the write event".format(
                        retry_count=i, max_count=retry_limit
                    ),
                    exc_info=True,
                )
                raise e
            else:
                crate_logger.warning(
                    "Attempt ({retry_count}) "
                    "out of ({max_count})".format(retry_count=i, max_count=retry_limit),
                    exc_info=True,
                )
    return cur_box_folder  # type: ignore[reportUnboundVariable]


def path_time_recurse_func(
    cur_path: Union[str, Path], inotify_wm: pyinotify.WatchManager = None
) -> Dict[str, float]:
    cur_data_map = {
        Path(cur_path)
        .resolve()
        .as_posix(): datetime.fromtimestamp(os.path.getmtime(cur_path))
        .astimezone(tz.UTC)
        .timestamp()
    }
    cur_path = Path(cur_path)
    for sub_cur_path in cur_path.iterdir():
        if inotify_wm:
            pass
            # wm.add_watch(sub_cur_path.as_posix(), mask, rec=True, auto_add=True)
        if sub_cur_path.is_file():
            cur_data_map[sub_cur_path.resolve().as_posix()] = (
                datetime.fromtimestamp(sub_cur_path.stat().st_mtime)
                .astimezone(dateutil.tz.tzutc())
                .timestamp()
            )
        else:
            if inotify_wm:
                pass
                # wm.add_watch(sub_cur_path.as_posix(), mask, rec=True, auto_add=True)
            cur_data_map.update(
                **path_time_recurse_func(sub_cur_path, inotify_wm=inotify_wm)
            )
    return cur_data_map


wm = pyinotify.WatchManager()
in_delete = getattr(pyinotify, "IN_DELETE")
in_delete_self = getattr(pyinotify, "IN_DELETE_SELF")
in_move_self = getattr(pyinotify, "IN_MOVE_SELF")
in_create = getattr(pyinotify, "IN_CREATE")
in_modify = getattr(pyinotify, "IN_MODIFY")
in_close_write = getattr(pyinotify, "IN_CLOSE_WRITE")
in_moved_to = getattr(pyinotify, "IN_MOVED_TO")
in_moved_from = getattr(pyinotify, "IN_MOVED_FROM")
all_events = getattr(pyinotify, "ALL_EVENTS")
mask = (
    in_delete | in_modify | in_close_write | in_moved_to | in_moved_from
)  # watched events



import json
import logging
import os
import time
from os import PathLike
from pathlib import Path
from typing import Union, Optional, List, Dict, Any

import redis
from boxsdk.object.file import File
from boxsdk.object.folder import Folder

from . import setup_logger

setup_logger()

crate_logger = logging.getLogger(__name__)


def redis_key(key: str) -> str:
    """

    :param key:
    :return:
    """
    return "diy_crate.version.{}".format(key)


def redis_path_for_object_id_key(key: PathLike):
    key_str = key.as_posix() if isinstance(key, Path) else key
    return f"diy_crate.path_for_object_id:{key_str}"


def redis_set(
    cache_client: redis.Redis,
    cloud_item: Union[File, Folder],
    last_modified_time: float,
    box_dir_path: PathLike,
    fresh_download: bool = False,
    folder: Union[str, Path] = None,
    sub_ids: Optional[List[str]] = None,
    parent_id: Optional[str] = None,
) -> Dict[str, Any]:
    """

    :param cache_client:
    :param cloud_item:
    :param last_modified_time:
    :param box_dir_path:
    :param fresh_download:
    :param folder:
    :param sub_ids:
    :param parent_id:
    :return:
    """
    key = redis_key(cloud_item.object_id)
    if "etag" not in cloud_item:
        cloud_item = cloud_item.get(fields=["etag", "path_collection", "name"])
    if folder:
        if isinstance(folder, Path):
            path = folder
        else:
            path = Path(folder)
    elif int(cloud_item["path_collection"]["total_count"]) > 1:
        path = Path(
            os.path.sep.join(
                [
                    folder["name"]
                    for folder in cloud_item["path_collection"]["entries"][1:]
                ]
            )
        )
    else:
        path = Path()
    path = Path(box_dir_path) / path
    item_info = {
        "fresh_download": fresh_download,
        "time_stamp": last_modified_time,
        "etag": cloud_item["etag"],
        "file_path": (path / cloud_item["name"]).as_posix(),
    }
    if sub_ids is not None:
        item_info["sub_ids"] = sub_ids
    if parent_id:
        item_info["parent_id"] = parent_id
    cache_client.set(key, json.dumps(item_info))
    last_save_time_stamp = int(time.time())
    cache_client.set("diy_crate.last_save_time_stamp", last_save_time_stamp)
    cache_client.set(
        redis_path_for_object_id_key((path / cloud_item["name"])), cloud_item.object_id
    )

    crate_logger.debug(f"Storing/updating info to redis:  {key=} {item_info=}")
    # assert redis_get(obj)
    return {key: item_info, "diy_crate.last_save_time_stamp": last_save_time_stamp}


def redis_get(cache_client, obj: Union[File, Folder]):
    """

    :param cache_client:
    :param obj:
    :return:
    """
    key = redis_key(obj.object_id)
    return json.loads(str(cache_client.get(key), encoding="utf-8", errors="strict"))


def id_for_file_path(cache_client, file_path):
    """

    :param cache_client:
    :param file_path:
    :return:
    """
    maybe_obj_id = cache_client.get(redis_path_for_object_id_key(file_path))
    if maybe_obj_id:
        return str(maybe_obj_id)
    for key in cache_client.keys("diy_crate.version.*"):
        value = json.loads(
            str(cache_client.get(key), encoding="utf-8", errors="strict")
        )
        if value.get("file_path") == file_path:
            return str(key, encoding="utf-8", errors="strict").split(".")[-1]


r_c = redis.StrictRedis()


def local_or_box_file_m_time_key_func(x: Union[PathLike, str], from_box: bool) -> str:
    return f"diy_crate.{'box' if from_box else 'local'}_mtime.{x}"


def local_or_box_file_m_time(x: PathLike, from_box: bool) -> float:
    return float(r_c.get(local_or_box_file_m_time_key_func(x, from_box)))
